from . import (  # noqa: F401
    basis,
    clebsch_gordan,
    cutoff,
    density,
    splines,
    utils,
)
from ._hypers import BadHyperParameters, convert_hypers, hypers_to_json  # noqa: F401
from .calculator_base import CalculatorBase  # noqa: F401

# don't forget to also update `featomic/torch/__init__.py` and
# `featomic/torch/calculators.py` when modifying this file
from .calculators import (
    AtomicComposition,
    LodeSphericalExpansion,
    NeighborList,
    SoapPowerSpectrum,
    SoapRadialSpectrum,
    SortedDistances,
    SphericalExpansion,
    SphericalExpansionByPair,
)
from .log import set_logging_callback  # noqa: F401
from .profiling import Profiler  # noqa: F401
from .status import FeatomicError  # noqa: F401
from .systems import IntoSystem, SystemBase  # noqa: F401
from .version import __version__  # noqa: F401


__all__ = [
    "AtomicComposition",
    "LodeSphericalExpansion",
    "NeighborList",
    "SoapPowerSpectrum",
    "SoapRadialSpectrum",
    "SortedDistances",
    "SphericalExpansion",
    "SphericalExpansionByPair",
]



"""
This module provides a convenience calculator for performing density auto-
correlations, i.e. the (iterative) CG tensor products of density (body
order 2) tensors.

This wraps the :py:class:`ClebschGordanProduct` class, handling the higher-level
metadata manipulation to produce the desired output tensors.
"""

from typing import List, Optional

from . import _dispatch
from ._backend import Device, DType, Labels, TensorMap, TorchModule, operations
from ._cg_product import ClebschGordanProduct


class DensityCorrelations(TorchModule):
    """
    Takes ``n_correlations`` of iterative CG tensor products of a density with itself to
    produce a density auto-correlation tensor of higher correlation order.

    The constructor computes and stores the CG coefficients. The :py:meth:`compute`
    method computes the auto-correlation by CG tensor product of the input density.
    """

    def __init__(
        self,
        *,
        n_correlations: int,
        max_angular: int,
        skip_redundant: bool = True,
        cg_backend: Optional[str] = None,
        arrays_backend: Optional[str] = None,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ):
        """
        :param n_correlations: :py:class:`int`, the number of iterative CG tensor
                products to perform.
        :param max_angular: :py:class:`int`, the maximum angular momentum to compute CG
            coefficients for. This must be large enough to perform the desired number of
            correlations on the density passed to the :py:meth:`compute` method.
        :param skip_redundant: :py:class:`bool`, whether to skip redundant CG
            combination steps without losing information in the output. Setting to
            ``True`` can save computation time, but does not preserve the norm of the
            output tensors.
        :param cg_backend: :py:class:`str`, the backend to use for the CG tensor
            product. If ``None``, the backend is automatically selected based on the
            arrays backend.
        :param arrays_backend: :py:class:`str`, the backend to use for array operations.
            If ``None``, the backend is automatically selected based on the environment.
            Possible values are "numpy" and "torch".
        :param dtype: the scalar type to use to store coefficients
        :param device: the computational device to use for calculations. This must be
            ``"cpu"`` if ``array_backend="numpy"``.
        """

        super().__init__()

        self._n_correlations = n_correlations
        self._cg_product = ClebschGordanProduct(
            max_angular=max_angular,
            cg_backend=cg_backend,
            keys_filter=_filter_redundant_keys if skip_redundant else None,
            arrays_backend=arrays_backend,
            dtype=dtype,
            device=device,
        )

    def compute(
        self,
        density: TensorMap,
        angular_cutoff: Optional[int] = None,
        selected_keys: Optional[Labels] = None,
    ) -> TensorMap:
        """
        Takes ``n_correlations`` of iterative CG tensor products of a density with
        itself, to generate density auto-correlations of arbitrary correlation order.

        .. math::

            \\rho^{\\nu=n_{corr} + 1} = \\rho^{\\nu=1}
                                        \\otimes \\rho^{\\nu=1} \\ldots \\otimes
                                        \\rho^{\\nu=1}

        where \\rho^{\\nu=1} is the input ``density`` of correlation order 1 (body order
        2), and \\rho^{\\nu=n_{corr} + 1} is the output density of correlation order
        ``n_correlations + 1``.

        The input ``density`` must have at least ``"o3_lambda"`` and  ``"o3_sigma"`` key
        dimensions, as well as a single components axis with the ``"o3_mu"`` dimension.
        Any other key dimensions present will be matched, while all properties
        dimensions will have their full tensor product computed.

        For instance, if the key dimensions of ``density`` are ``"o3_lambda"``,
        ``"o3_sigma"``, ``"center_type"``, and ``"neighbor_type"``, pairs of blocks will
        only be combined if they have both the same ``"center_type"`` and
        ``"neighbor_type"``. To fully correlate all combinations of ``"neighbor_type"``,
        for example, this key dimension should be moved to properties first, using
        :py:meth:`TensorMap.keys_to_properties`.

        Before performing any correlations, the properties dimensions of ``density`` are
        modified to carry a "_1" suffix. At each iteration, the dimension names of the
        copy of the density being correlated are incremented by one each time.

        ``selected_keys`` can be passed to select the keys to compute on the final
        iteration. If ``None``, all keys are computed. To limit the maximum angular
        momenta to compute on **intermediate** iterations, pass ``angular_cutoff``.

        If ``angular_cutoff`` and ``selected_keys`` are both passed, ``angular_cutoff``
        is ignored on the final iteration.

        :param density: :py:class:`TensorMap`, the input density tensor of correlation
            order 1.
        :param angular_cutoff: :py:class:`int`, the maximum angular momentum to compute
            output blocks for at each iteration. If ``selected_keys`` is passed, this
            parameter is applied to all intermediate (i.e. prior to the final)
            iterations. If ``selected_keys`` is not passed, this parameter is applied
            globally to all iterations. If ``None``, all angular momenta are computed.
        :param selected_keys: :py:class:`Labels`, the keys to compute on the final
            iteration. If ``None``, all keys are computed. Subsets of key dimensions can
            be passed to compute output blocks that match in these dimensions.

        :return: :py:class:`TensorMap`, the output density auto-correlation tensor of
            correlation order ``n_correlations + 1``.
        """
        return self._density_correlations(
            density,
            angular_cutoff=angular_cutoff,
            selected_keys=selected_keys,
            compute_metadata=False,
        )

    def forward(
        self,
        density: TensorMap,
        angular_cutoff: Optional[int] = None,
        selected_keys: Optional[Labels] = None,
    ) -> TensorMap:
        """
        Calls the :py:meth:`compute` method.

        This is intended for :py:class:`torch.nn.Module` compatibility, and should be
        ignored in pure Python mode.

        See :py:meth:`compute` for a full description of the parameters.
        """
        return self.compute(
            density, angular_cutoff=angular_cutoff, selected_keys=selected_keys
        )

    def compute_metadata(
        self,
        density: TensorMap,
        angular_cutoff: Optional[int] = None,
        selected_keys: Optional[Labels] = None,
    ) -> TensorMap:
        """
        Returns the metadata-only :py:class:`TensorMap` that would be output by the
        function :py:meth:`compute` for the same calculator under the same settings,
        without performing the actual Clebsch-Gordan tensor products.

        See :py:meth:`compute` for a full description of the parameters.
        """
        return self._density_correlations(
            density,
            angular_cutoff=angular_cutoff,
            selected_keys=selected_keys,
            compute_metadata=True,
        )

    def _density_correlations(
        self,
        density: TensorMap,
        angular_cutoff: Optional[int],
        selected_keys: Optional[Labels],
        compute_metadata: bool,
    ) -> TensorMap:
        """
        Computes the density auto-correlations.
        """
        # Prepare the tensors. Before the first iteration, both the 'left' and 'right'
        # densities should have a "_1" suffix in the property names. The property names
        # of the 'right' density is incremented by one at the start of each iteration.
        density_correlations = _increment_property_names(density, 1)
        density = _increment_property_names(density, 1)

        # Perform iterative CG tensor products
        new_lambda_names: List[str] = []
        for i_correlation in range(self._n_correlations):
            # Increment the density property dimension names
            density = _increment_property_names(density, 1)

            # Define new key dimension names for tracking intermediate correlations. The
            # following conditional clause is a matter of convention. "l" indices track
            # blocks of nu=1 densities, and "k" indices track blocks originating from
            # higher order tensors.
            if i_correlation == 0:
                o3_lambda_1_new_name: str = f"l_{i_correlation + 1}"
            else:
                o3_lambda_1_new_name: str = f"k_{i_correlation + 1}"
            o3_lambda_2_new_name: str = f"l_{i_correlation + 2}"
            new_lambda_names.extend([o3_lambda_1_new_name, o3_lambda_2_new_name])

            # Define the selected keys for the current iteration, applying the
            # ``angular_cutoff`` if specified.
            if (
                i_correlation == self._n_correlations - 1 and selected_keys is None
            ) or (i_correlation < self._n_correlations - 1):
                if angular_cutoff is None:
                    selected_keys_iteration = None
                else:
                    selected_keys_iteration: Labels = Labels(
                        names=["o3_lambda"],
                        values=_dispatch.int_range_like(
                            0, angular_cutoff + 1, like=density.keys.values
                        ).reshape(-1, 1),
                    )
            else:
                assert i_correlation == self._n_correlations - 1
                assert selected_keys is not None
                selected_keys_iteration = selected_keys

            # Compute CG tensor product
            if compute_metadata:
                density_correlations = self._cg_product.compute_metadata(
                    density_correlations,
                    density,
                    o3_lambda_1_new_name,
                    o3_lambda_2_new_name,
                    selected_keys=selected_keys_iteration,
                )
            else:
                density_correlations = self._cg_product.compute(
                    density_correlations,
                    density,
                    o3_lambda_1_new_name,
                    o3_lambda_2_new_name,
                    selected_keys=selected_keys_iteration,
                )

        return density_correlations


def _filter_redundant_keys(keys: Labels) -> List[int]:
    """
    Filter redundant keys from the ``keys`` to only keep keys where the l values are
    sorted (i.e. l1 <= l2 <= ... <= ln). These are redundant when handling
    auto-correlations of a density.
    """
    # Infer the correlation order of the output TensorMap by the length of the "l" list
    # present in the keys
    nu_target, present = 1, True
    while present:
        if f"l_{nu_target + 1}" in keys.names:
            nu_target += 1
        else:
            present = False

    # As this function is only valid for the output of a density auto-correlation, we
    # assert the "k" list is present in the keys
    assert all([f"k_{k}" in keys.names for k in range(2, nu_target)])

    # Now find the column idxs of the "l" values in the "l" list
    l_list_idxs = [
        keys.names.index(f"l_{o3_lambda}") for o3_lambda in range(1, nu_target + 1)
    ]

    # Identify keys of redundant blocks and remove them
    keys_to_keep: List[int] = []
    for key_idx in range(len(keys)):
        key = keys.entry(key_idx)

        # Get the list of "l" values
        l_list_values = _dispatch.to_int_list(key.values[l_list_idxs])

        # Keep this key if the l list is already sorted
        if _dispatch.all(
            _dispatch.int_array_like(l_list_values, like=key.values)
            == _dispatch.int_array_like(sorted(l_list_values), like=key.values)
        ):
            keys_to_keep.append(key_idx)

    return keys_to_keep


def _increment_property_names(
    tensor: TensorMap, increment_by: Optional[int] = None
) -> TensorMap:
    """
    Increments all the property dimension names of the input :py:class:`TensorMap`
    ``tensor`` by ``increment_by`` and returns the resulting :py:class:`TensorMap`.

    Any property dimension that is already suffixed by "_{x}", where "x" is an integer,
    will have the integer incremented by ``increment_by``. If the property dimension is
    not suffixed by "_{x}", then "_0" is implied and then this is incremented.

    All property dimensions are assumed to have the format "{property_name}_{x}". The
    only exception are property names that indicate type, and have the format
    "{name}_type". In these cases, the numeric index precedes the "_type" suffix.

    For instance, if the input tensor has property names ["n", "neighbor_type"], the
    property names of the output of ``_increment_property_names(tensor, 2)`` will
    be ["n_2", "neighbor_2_type"]. If the input tensor has property names ["n",
    "neighbor_1_type"], the output property names of the same function call will be
    ["n_2", "neighbor_3_type"].

    :param tensor: :py:class:`TensorMap`, the input tensor.
    :param increment_by: :py:class:`int`, the integer amount to increment the numeric
        suffixes by. Must be a non-negative integer. If ``None``, the default value of 1
        is used.

    :return: :py:class:`TensorMap`, the input tensor with all property dimension names
        incremented by ``increment_by``, according to the rules described above.
    """
    if increment_by is None:
        increment_by = 1
    if increment_by < 0:
        raise ValueError("`increment_by` must be a non-negative integer.")

    for name in tensor.property_names:
        tensor = operations.rename_dimension(
            tensor,
            "properties",
            name,
            _increment_numeric_suffix(name, increment_by),
        )

    return tensor


def _increment_numeric_suffix(name: str, increment_by: Optional[int] = 1) -> str:
    """
    Takes a string name that is suffix by "_{x}", where "x" is an integer,
    and increments the integer by ``increment_by``. Returns the new name.

    If the string is not suffixed by "_{x}", then "_1" is appended to the string.

    The only special cases are names suffixed by "_type", in which case the numeric
    index is inserted and incremented before the "_type" suffix.

    Examples:
        - fn("n", 1) -> "n_1"
        - fn("n", 0) -> "n_0"
        - fn("n_1", 1) -> "n_2"
        - fn("center_type", 3) -> "center_3_type"
        - fn("first_atom_1_type") -> "first_atom_2_type"
        - fn("center_type_1_type", 7) -> "center_type_8_type"
    """
    if not isinstance(increment_by, int):
        raise ValueError("`increment_by` must be an integer.")

    if name.endswith("type"):
        # Special name case: remove the "_type" suffix and add it back on at the end
        prefix: str = name[: name.rfind("_type")]
        suffix: str = "_type"
    else:
        prefix: str = name
        suffix: str = ""

    number: str = ""
    if prefix.rfind("_") == -1:  # no suffix present
        number = "0"
    else:
        number = prefix[prefix.rfind("_") + 1 :]
        if number.isdigit():  # number found
            prefix = prefix[: prefix.rfind("_")]
        else:
            number = "0"

    return prefix + "_" + str(int(number) + increment_by) + suffix



"""
Module for computing the Clebsch-Gordan tensor product between two arbitrary
:py:class:`TensorMap` objects.
"""

from typing import Callable, List, Optional

import numpy as np

from . import _coefficients, _dispatch, _utils
from ._backend import (
    Device,
    DType,
    Labels,
    TensorBlock,
    TensorMap,
    TorchModule,
    TorchScriptClass,
    torch_jit_is_scripting,
)


try:
    import torch
except ImportError:
    pass


class ClebschGordanProduct(TorchModule):
    """
    A general class for computing the Clebsch-Gordan (CG) tensor product between two
    arbitrary :py:class:`TensorMap`.

    The constructor computes and stores the CG coefficients. The :py:meth:`compute`
    method computes the CG tensor product between two tensors.
    """

    def __init__(
        self,
        *,
        max_angular: int,
        cg_backend: Optional[str] = None,
        keys_filter: Optional[Callable[[Labels], List[int]]] = None,
        arrays_backend: Optional[str] = None,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ):
        """
        :param max_angular: :py:class:`int`, the maximum angular momentum to compute CG
            coefficients for.
        :param cg_backend: :py:class:`str`, the backend to use for the CG tensor
            product. If ``None``, the backend is automatically selected based on the
            arrays backend ("numpy" when importing this class from ``featomic.utils``,
            and "torch" when importing from ``featomic.torch.utils``).
        :param keys_filter: A function to remove more keys from the output. This is
            applied after any user-provided ``key_selection`` in :py:meth:`compute`.
            This function should take one argument ``keys: Labels``, and return the
            indices of keys to keep.
        :param arrays_backend: :py:class:`str`, the backend to use for array operations.
            If ``None``, the backend is automatically selected based on the environment.
            Possible values are "numpy" and "torch".
        :param dtype: the scalar type to use to store coefficients
        :param device: the computational device to use for calculations. This must be
            ``"cpu"`` if ``array_backend="numpy"``.
        """

        super().__init__()

        # Assign the arrays backend
        if arrays_backend is None:
            if torch_jit_is_scripting():
                arrays_backend = "torch"
            else:
                if isinstance(Labels, TorchScriptClass):
                    arrays_backend = "torch"
                else:
                    arrays_backend = "numpy"
        elif arrays_backend == "numpy":
            if torch_jit_is_scripting():
                raise ValueError(
                    "Module is torch scripted but 'numpy' was given as `arrays_backend`"
                )
            arrays_backend = "numpy"
        elif arrays_backend == "torch":
            # all good
            pass
        else:
            raise ValueError(
                f"Unknown `arrays_backend` {arrays_backend}."
                "Only 'numpy' and 'torch' are supported."
            )

        if dtype is None:
            if arrays_backend == "torch":
                dtype = torch.get_default_dtype()
            elif arrays_backend == "numpy":
                dtype = np.float64

        if device is None:
            if arrays_backend == "torch":
                device = torch.get_default_device()
            elif arrays_backend == "numpy":
                device = "cpu"

        if arrays_backend == "numpy":
            if device != "cpu":
                raise ValueError("`device` can only be 'cpu' for numpy arrays")

        # Assign the CG backend
        if cg_backend is None:
            if arrays_backend == "torch":
                self._cg_backend = "python-dense"
            else:
                self._cg_backend = "python-sparse"
        else:
            self._cg_backend = cg_backend

        if max_angular < 0:
            raise ValueError(
                f"Given `max_angular={max_angular}` negative. Must be greater equal 0."
            )
        self._max_angular = max_angular

        # Compute the CG coefficients
        self._cg_coefficients = _coefficients.calculate_cg_coefficients(
            lambda_max=self._max_angular,
            cg_backend=self._cg_backend,
            arrays_backend=arrays_backend,
            dtype=dtype,
            device=device,
        )

        if keys_filter is None:
            self._keys_filter = _keys_filter_noop
        else:
            self._keys_filter = keys_filter

    def compute(
        self,
        tensor_1: TensorMap,
        tensor_2: TensorMap,
        o3_lambda_1_new_name: str,
        o3_lambda_2_new_name: str,
        selected_keys: Optional[Labels] = None,
    ) -> TensorMap:
        """
        Computes the Clebsch-Gordan (CG) tensor product between ``tensor_1`` and
        ``tensor_2``.

        This function assumes the metadata of ``tensor_1`` and ``tensor_2`` has been
        modified to be compatible for the CG tensor product, according to the following
        rules:

        - both ``tensor_1`` and ``tensor_2`` must have key dimensions ``"o3_lambda"``
          and ``"o3_sigma"`` as these are used to determine the symmetry of output
          blocks upon combination;
        - ``o3_lambda_1_new_name`` and ``o3_lambda_2_new_name`` define the output key
          dimension names that store the ``"o3_lambda"`` values of the blocks combined
          from ``tensor_1`` and ``tensor_2`` respectively;
        - any other key dimensions that have equivalent names in both ``tensor_1`` and
          ``tensor_2`` will be matched, such that only blocks that have equal values in
          these dimensions will be combined;
        - any other named key dimensions that are present in ``tensor_1`` but not in
            ``tensor_2``, and vice versa, will have the full product computed;
        - ``tensor_1`` and ``tensor_2`` must have a single component axis with a single
          key dimension named ``"o3_mu"``. ``tensor_1`` and ``tensor_2`` may have
          different samples, but all samples in the tensor with fewer samples dimensions
          must be present in the samples of the other tensor.

        A :py:class:`Labels` object can be passed in ``selected_keys`` to select
        specific keys to compute. The full set of output keys are computed, then are
        filtered to match the key dimensions passed in ``selected_keys``.

        This parameter can be used to match the output tensor to a given target basis
        set definition, and/or enhance performance by limiting the combinations
        computed.

        For instance, passing just the ``"o3_lambda"`` key dimension with a range of
        values ``0, ..., max_angular`` can be used to perform angular selection,
        limiting the maximum angular momentum of the output blocks.

        Note that using ``selected_keys`` to perform any kind of selection will reduce
        the information content of the output tensor. This may be important if the
        returned tensor is used in further CG tensor products.

        :param tensor_1: The first :py:class:`TensorMap`, containing data with SO(3)
            character.
        :param tensor_2: The first :py:class:`TensorMap`, containing data with SO(3)
            character.
        :param o3_lambda_1_new_name: :py:class:`str`, the name of the output key
            dimension that stores the ``"o3_lambda"`` values of the blocks combined from
            ``tensor_1``.
        :param o3_lambda_2_new_name: :py:class:`str`, the name of the output key
            dimension that stores the ``"o3_lambda"`` values of the blocks combined from
            ``tensor_2``.
        :param selected_keys: :py:class:`Labels`, the keys to compute on the final
            iteration. If ``None``, all keys are computed. Subsets of key dimensions can
            be passed to compute output blocks that match in these dimensions.

        :return: A :py:class:`TensorMap` containing the Clebsch-Gordan tensor product of
            ``tensor_1`` and ``tensor_2``.
        """

        return self._cg_tensor_product(
            tensor_1,
            tensor_2,
            o3_lambda_1_new_name=o3_lambda_1_new_name,
            o3_lambda_2_new_name=o3_lambda_2_new_name,
            selected_keys=selected_keys,
            compute_metadata=False,
        )

    def forward(
        self,
        tensor_1: TensorMap,
        tensor_2: TensorMap,
        o3_lambda_1_new_name: str,
        o3_lambda_2_new_name: str,
        selected_keys: Optional[Labels] = None,
    ) -> TensorMap:
        """
        Calls the :py:meth:`ClebschGordanProduct.compute` function.

        This is intended for :py:class:`torch.nn.Module` compatibility, and should be
        ignored in pure Python mode.

        See :py:meth:`compute` for a full description of the parameters.
        """
        return self.compute(
            tensor_1,
            tensor_2,
            o3_lambda_1_new_name=o3_lambda_1_new_name,
            o3_lambda_2_new_name=o3_lambda_2_new_name,
            selected_keys=selected_keys,
        )

    def compute_metadata(
        self,
        tensor_1: TensorMap,
        tensor_2: TensorMap,
        o3_lambda_1_new_name: str,
        o3_lambda_2_new_name: str,
        selected_keys: Optional[Labels] = None,
    ) -> TensorMap:
        """
        Returns the metadata-only :py:class:`TensorMap` that would be output by the
        function :py:meth:`compute` for the same calculator under the same settings,
        without performing the actual Clebsch-Gordan tensor products.

        See :py:meth:`compute` for a full description of the parameters.
        """
        return self._cg_tensor_product(
            tensor_1,
            tensor_2,
            o3_lambda_1_new_name=o3_lambda_1_new_name,
            o3_lambda_2_new_name=o3_lambda_2_new_name,
            selected_keys=selected_keys,
            compute_metadata=True,
        )

    def _cg_tensor_product(
        self,
        tensor_1: TensorMap,
        tensor_2: TensorMap,
        o3_lambda_1_new_name: str,
        o3_lambda_2_new_name: str,
        selected_keys: Optional[Labels],
        compute_metadata: bool,
    ) -> TensorMap:
        """
        Computes the Clebsch-Gordan tensor product between ``tensor_1`` and
        ``tensor_2``.

        Executes the following steps:

            1. Checks the metadata of ``tensor_1`` and ``tensor_2`` to ensure that they
               are compatible for the Clebsch-Gordan tensor product. Ensures neither has
               gradients (not currently supported).

            2. Computes the full product of the two keys, accounting for CG combination
               rules, and matching key dimensions that have the same name.

            3. Applies the key selection to remove keys that do not need computing. This
               includes both ``selected_keys`` and any user-provided ``keys_filter``.

            4. For each key pair to be combined (and corresponding output angular
               channel), computes the CG tensor product on the block values.

            5. Returns the resulting tensor in :py:class:`TensorMap` format.
        """
        # 1. Check the inputs
        self._check_inputs(tensor_1, tensor_2, selected_keys)

        # 2. Compute the full product of the two keys
        output_keys, combinations = _utils._compute_output_keys(
            tensor_1.keys, tensor_2.keys, o3_lambda_1_new_name, o3_lambda_2_new_name
        )

        # 3. a) Apply key selections
        if selected_keys is not None:
            selected_idx = output_keys.select(selected_keys)
            combinations = [combinations[i] for i in selected_idx]
            output_keys = Labels(
                names=output_keys.names, values=output_keys.values[selected_idx]
            )

        # 3. b) Apply key filter
        keys_to_keep = self._keys_filter(output_keys)
        output_keys = Labels(output_keys.names, output_keys.values[keys_to_keep])
        combinations = [combinations[i] for i in keys_to_keep]

        # Group keys such that all o3_lambda values produced by the same 2 blocks are
        # together. This allows the block combination and CG product to be separated and
        # thus not unnecessarily repeated.
        output_keys, combinations = _utils._group_combinations_of_same_blocks(
            output_keys, combinations
        )

        # 4. Do the CG tensor product for each block combination
        output_blocks: List[TensorBlock] = []
        for combination in combinations:
            output_blocks.extend(
                _utils.cg_tensor_product_blocks(
                    tensor_1.block(combination.first),
                    tensor_2.block(combination.second),
                    o3_lambdas=combination.o3_lambdas,
                    cg_coefficients=self._cg_coefficients,
                    cg_backend="metadata" if compute_metadata else self._cg_backend,
                )
            )

        # 5. Build and return the resulting TensorMap
        return TensorMap(output_keys, output_blocks)

    def _check_inputs(
        self,
        tensor_1: TensorMap,
        tensor_2: TensorMap,
        selected_keys: Optional[Labels] = None,
    ) -> None:
        """
        Checks the metadata of ``tensor_1`` and ``tensor_2`` to ensure that they are
        compatible for the Clebsch-Gordan tensor product.

        This checks that:

            - Both tensors have "o3_lambda" and "o3_sigma" key dimensions.
            - Both tensors have a single component axis with a single dimension named
              "o3_mu".
            - The sum of maximum "o3_lambda" values in each tensor is not greater than
              the maximum angular momentum ``self._max_angular`` used to calculate the
              CG coefficients in the constructor.
            - There is no intersection of property names between the two tensors.
        """
        # Check symmetry dimensions
        for tensor in [tensor_1, tensor_2]:
            if "o3_lambda" not in tensor.keys.names:
                raise ValueError(
                    "input tensors must have a named key dimension `'o3_lambda`"
                )
            if "o3_sigma" not in tensor.keys.names:
                raise ValueError(
                    "input tensors must have a named key dimension `'o3_sigma`"
                )
            if not tensor.component_names == ["o3_mu"]:
                raise ValueError(
                    "input tensors must have a single component"
                    " axis with a single named dimensions `'o3_mu'`"
                )

        # Check maximum angular momentum
        max_angular: int = int(
            _dispatch.max(tensor_1.keys.column("o3_lambda"))
            + _dispatch.max(tensor_2.keys.column("o3_lambda"))
        )
        if selected_keys is not None:
            # Update `max_angular` if angular selection is being performed via
            # `selected_keys`
            if "o3_lambda" in selected_keys.names:
                # Check max of 'o3_lambda' column is valid
                for lam in selected_keys.column("o3_lambda"):
                    if lam > self._max_angular:
                        raise ValueError(
                            "the maximum angular momentum value found in key dimension"
                            " `'o3_lambda'` in `selected_keys` exceeds"
                            f" `max_angular={self._max_angular}`"
                            " used to calculate the CG coefficients in the constructor."
                        )
                max_angular = min(
                    max_angular, int(_dispatch.max(selected_keys.column("o3_lambda")))
                )
            if "o3_sigma" in selected_keys.names:
                # Check 'o3_sigma' column is valid
                for o3_sigma in selected_keys.column("o3_sigma"):
                    if int(o3_sigma) not in [-1, 1]:
                        raise ValueError(
                            "key dimension `'o3_sigma'` in `selected_keys` must only"
                            " contain values of -1 or 1."
                        )
        if max_angular > int(self._max_angular):
            raise ValueError(
                f"the maximum angular momentum `{max_angular}` required by the"
                " input tensors (and perhaps lowered by angular selection via"
                f" `selected_keys`) exceeds the `max_angular={self._max_angular}`"
                " used to calculate the CG coefficients in the constructor."
            )

        # Check property names
        for property_name in tensor_1.property_names:
            if property_name in tensor_2.property_names:
                raise ValueError(
                    f"property name `{property_name}` present in both input tensors."
                    " As all property dimensions are combined, they must have"
                    " different names in the two tensors. Use"
                    " `metatensor.rename_dimension` and try again."
                )

        # Check no gradients as not currently supported.
        contains_gradients_1 = all(
            [len(list(block.gradients())) > 0 for _, block in tensor_1.items()]
        )
        contains_gradients_2 = all(
            [len(list(block.gradients())) > 0 for _, block in tensor_2.items()]
        )
        if contains_gradients_1 or contains_gradients_2:
            raise NotImplementedError(
                "Clebsch Gordan combinations with gradients not yet implemented."
                " Use `metatensor.remove_gradients` to remove gradients from the input"
                " tensors."
            )


def _keys_filter_noop(keys: Labels):
    """A ``key_filter`` for ``ClebschGordanProduct`` that does nothing"""
    return [i for i in range(len(keys))]



import functools
from typing import Callable, Optional

import numpy as np


try:
    import scipy.integrate
    import scipy.special

    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False

from .basis import ExpansionBasis, Explicit, RadialBasis
from .cutoff import Cutoff
from .density import AtomicDensity, DiracDelta, Gaussian, SmearedPowerLaw


class Spline:
    """
    `Cubic Hermite splines`_ implementation

    .. _Cubic Hermit splines: https://en.wikipedia.org/wiki/Cubic_Hermite_spline
    """

    def __init__(self, start, stop):
        self.start = float(start)
        self.stop = float(stop)

        self.positions = None
        self.values = None
        self.derivatives = None

    def __len__(self):
        if self.positions is None:
            return 0
        else:
            return len(self.positions)

    def compute(
        self,
        positions: np.ndarray,
        n: Optional[int] = None,
        *,
        derivative=False,
    ) -> np.ndarray:
        """
        Evaluate the spline at the given ``positions``, optionally evaluating only
        the value for the ``n``'th splined function.
        """

        if self.positions is None:
            raise ValueError("you must add points to the spline before evaluating it")

        if n is None:
            n = ...

        x = positions
        delta_x = self.positions[1] - self.positions[0]
        k = (np.floor(x / delta_x)).astype(np.int64)

        t = (x - k * delta_x) / delta_x
        t_2 = t**2
        t_3 = t**3

        h00 = 2.0 * t_3 - 3.0 * t_2 + 1.0
        h10 = t_3 - 2.0 * t_2 + t
        h01 = -2.0 * t_3 + 3.0 * t_2
        h11 = t_3 - t_2

        p_k = self.values[k, n]
        p_k_1 = self.values[k + 1, n]

        m_k = self.derivatives[k, n]
        m_k_1 = self.derivatives[k + 1, n]

        new_shape = (-1,) + (1,) * (len(p_k.shape) - 1)
        h00 = h00.reshape(new_shape)
        h10 = h10.reshape(new_shape)
        h01 = h01.reshape(new_shape)
        h11 = h11.reshape(new_shape)

        if derivative:
            d_h00_dt = 6.0 * (t_2 - t)
            d_h10_dt = 3.0 * t_2 - 4.0 * t + 1.0
            d_h01_dt = -d_h00_dt
            d_h11_dt = 3.0 * t_2 - 2.0 * t

            dx_dt = 1.0 / delta_x

            return (
                d_h00_dt * p_k * dx_dt
                + d_h10_dt * m_k
                + d_h01_dt * p_k_1 * dx_dt
                + d_h11_dt * m_k_1
            )

        else:
            return h00 * p_k + h10 * delta_x * m_k + h01 * p_k_1 + h11 * delta_x * m_k_1

    def add_points(self, positions, values, derivatives):
        """Add points to the spline

        :param positions: positions of all points
        :param values: values of the splined functions at the points positions
        :param derivatives: derivative of the splined functions at the points positions
        """
        positions = np.asarray(positions)
        values = np.asarray(values)
        derivatives = np.asarray(derivatives)

        if not np.all(np.isfinite(positions)):
            raise ValueError(
                "new spline points `positions` contains NaN/infinity, "
                "numerical integrals are not converging"
            )

        if not np.all(np.isfinite(values)):
            raise ValueError(
                "new spline points `values` contains NaN/infinity, "
                "numerical integrals are not converging"
            )

        if not np.all(np.isfinite(derivatives)):
            raise ValueError(
                "new spline points `derivatives` contains NaN/infinity, "
                "numerical integrals are not converging"
            )

        assert len(positions.shape) == 1
        assert np.min(positions) >= self.start
        assert np.max(positions) <= self.stop

        assert values.shape == derivatives.shape
        assert values.shape[0] == len(positions)
        assert derivatives.shape[0] == len(positions)

        if self.values is not None:
            assert values.shape[1:] == self.values.shape[1:]
            assert derivatives.shape[1:] == self.derivatives.shape[1:]

            positions = np.concatenate([self.positions, positions], axis=0)
            values = np.concatenate([self.values, values], axis=0)
            derivatives = np.concatenate([self.derivatives, derivatives], axis=0)

        sort_indices = np.argsort(positions, axis=0)
        self.positions = positions[sort_indices]
        self.values = values[sort_indices]
        self.derivatives = derivatives[sort_indices]

        assert len(self.values.shape) == 2

    @staticmethod
    def with_accuracy(
        start: float,
        stop: float,
        values_fn: Callable[[np.ndarray], np.ndarray],
        derivatives_fn: Callable[[np.ndarray], np.ndarray],
        accuracy: float,
    ) -> "Spline":
        """
        Create a :py:class:`Spline` using any set of functions defined within the
        ``start, stop`` interval. The same spline points will be used for all functions,
        and more will be added until either the relative error or the absolute error
        fall below the requested accuracy on average across all functions.

        The functions are specified via ``values_fn`` and ``derivatives_fn``. These must
        take the positions of new spline points as their input, and they must output a
        numpy array containing the full set of function evaluated at these points.
        """
        # initialize spline with 11 points
        spline = Spline(start=start, stop=stop)
        positions = np.linspace(start, stop, 11)
        spline.add_points(positions, values_fn(positions), derivatives_fn(positions))

        while True:
            n_intermediate_positions = len(spline) - 1

            if n_intermediate_positions >= 50000:
                raise ValueError(
                    "Maximum number of spline points reached. \
                    There might be a problem with the functions to be splined"
                )

            half_step = (spline.positions[1] - spline.positions[0]) / 2
            intermediate_positions = np.linspace(
                start + half_step, stop - half_step, n_intermediate_positions
            )

            estimated_values = spline.compute(intermediate_positions)
            new_values = values_fn(intermediate_positions)

            mean_absolute_error = np.mean(np.abs(estimated_values - new_values))
            with np.errstate(divide="ignore"):  # Ignore divide-by-zero warnings
                relative_error = np.abs((estimated_values - new_values) / (new_values))
                mean_relative_error = np.mean(
                    # exclude points where the denominator is (almost) zero
                    relative_error[np.where(np.abs(new_values) > 1e-16)]
                )

            if mean_absolute_error < accuracy or mean_relative_error < accuracy:
                break

            new_derivatives = derivatives_fn(intermediate_positions)

            spline.add_points(intermediate_positions, new_values, new_derivatives)

        return spline


class SplinedRadialBasis(RadialBasis):
    """
    Radial basis based on a spline. This is mainly intended to be used to transfer hyper
    parameters to the native code, but can also be used to check the exact shape of the
    splined radial basis.
    """

    def __init__(
        self,
        *,
        spline: Spline,
        max_radial: int,
        radius: float,
        lode_center_contribution: Optional[np.ndarray] = None,
    ):
        super().__init__(max_radial=max_radial, radius=radius)
        self.spline = spline
        self.lode_center_contribution = lode_center_contribution

    def get_hypers(self):
        hypers = {
            "type": "Tabulated",
            "points": [
                {
                    "position": float(p),
                    "values": v.tolist(),
                    "derivatives": d.tolist(),
                }
                for p, v, d in zip(
                    self.spline.positions,
                    self.spline.values,
                    self.spline.derivatives,
                )
            ],
        }

        if self.lode_center_contribution is not None:
            hypers["center_contribution"] = self.lode_center_contribution.tolist()

        return hypers

    def compute_primitive(
        self, positions: np.ndarray, n: int, *, derivative: bool
    ) -> np.ndarray:
        return self.spline.compute(positions, n, derivative=derivative)


def _spherical_bessel_scaled(ell, z):
    """Compute ``exp(-z) * spherical_in(ell, z)``, avoiding overflow in
    ``spherical_in`` for large values of ``z``"""
    one_over_z = np.divide(1.0, z, out=np.zeros_like(z), where=z != 0)

    result = np.sqrt(0.5 * np.pi * one_over_z) * scipy.special.ive(ell + 0.5, z)

    if ell == 0:
        return np.where(z == 0, 1.0, result)
    else:
        return result


class SoapSpliner:
    """Compute an explicit spline of the radial integral for SOAP calculators.

    This allows a great deal of customization in the radial basis function used (any
    child class of :py:class:`RadialBasis`) and atomic density (any child class of
    :py:class:`AtomicDensity`). This way, users can define custom densities and/or
    basis, and use them with any of the SOAP calculators. For more information about the
    radial integral, you can refer to :ref:`this document <radial-integral>`.

    This class should be used only in combination with SOAP calculators like
    :class:`featomic.SphericalExpansion` or :class:`featomic.SoapPowerSpectrum`. For
    k-space spherical expansions use :class:`LodeSpliner`.

    If ``density`` is either :class:`featomic.density.Delta` or
    :class:`featomic.density.Gaussian` the radial integral will be partly solved
    analytically, for faster and more stable evaluation.

    Example
    -------

    First let's define the hyper parameters for the spherical expansions. It is
    important to note that only class-based hyper parameters are supported (in
    opposition to ``dict`` based hyper parameters).

    >>> import featomic

    >>> cutoff = featomic.cutoff.Cutoff(radius=2, smoothing=None)
    >>> density = featomic.density.Gaussian(width=1.0)
    >>> basis = featomic.basis.TensorProduct(
    ...     max_angular=4,
    ...     radial=featomic.basis.Gto(max_radial=5, radius=2),
    ...     # use a reduced ``accuracy`` of ``1e-3`` (the default is ``1e-8``)
    ...     # to speed up calculations.
    ...     spline_accuracy=1e-3,
    ... )

    From here we can initialize the spliner instance

    >>> spliner = SoapSpliner(cutoff=cutoff, density=density, basis=basis)

    You can then give the result of :meth:`SoapSpliner.get_hypers()` directly the
    calculator:

    >>> calculator = featomic.SphericalExpansion(**spliner.get_hypers())

    .. seealso::
        :class:`LodeSpliner` for a spliner class that works with
        :class:`featomic.LodeSphericalExpansion`
    """

    def __init__(
        self,
        *,
        cutoff: Cutoff,
        density: AtomicDensity,
        basis: ExpansionBasis,
        n_spline_points: Optional[int] = None,
    ):
        """
        :param cutoff: description of the local atomic environment, defined by a
            spherical cutoff
        :param density: atomic density that should be expanded for all neighbors in the
            local atomic environment
        :param basis: basis function to use to expand the neighbors' atomic density.
        :param n_spline_points: number of spline points to use. If ``None``, points will
            be added to the spline until the accuracy is at least the requested
            ``basis.spline_accuracy``.
        """
        if not HAS_SCIPY:
            raise ImportError("SoapSpliner class requires scipy")

        if not isinstance(cutoff, Cutoff):
            raise TypeError(f"`cutoff` should be a `Cutoff` object, got {type(cutoff)}")

        if not isinstance(density, AtomicDensity):
            raise TypeError(
                f"`density` should be an `AtomicDensity` object, got {type(density)}"
            )

        if not isinstance(basis, ExpansionBasis):
            raise TypeError(
                f"`basis` should be an `ExpansionBasis object, got {type(basis)}"
            )

        self.cutoff = cutoff
        self.density = density
        self.basis = basis

        self.n_spline_points = n_spline_points

    def get_hypers(self):
        """Get the SOAP hyper-parameters for the splined basis and density."""
        # This class works by computing a spline for the radial integral, and then using
        # this spline as a basis in combination with a DiracDelta density.
        hypers = {
            "cutoff": self.cutoff,
            "density": DiracDelta(
                center_atom_weight=self.density.center_atom_weight,
                scaling=self.density.scaling,
            ),
        }

        def generic_values_fn(positions, radial_size, angular, orthonormalization):
            integrals = np.vstack(
                [
                    self._radial_integral(radial, angular, positions, derivative=False)
                    for radial in range(radial_size)
                ]
            )
            return (orthonormalization @ integrals).T

        def generic_derivatives_fn(positions, radial_size, angular, orthonormalization):
            integrals = np.vstack(
                [
                    self._radial_integral(radial, angular, positions, derivative=True)
                    for radial in range(radial_size)
                ]
            )
            return (orthonormalization @ integrals).T

        # We transform whatever basis the user provided to an "explicit" basis, since
        # the radial integral will be different for different angular channels.
        by_angular = {}
        for angular in self.basis.angular_channels():
            radial_basis = self.basis.radial_basis(angular)
            orthonormalization = radial_basis._get_orthonormalization_matrix()

            values_fn = functools.partial(
                generic_values_fn,
                radial_size=radial_basis.size,
                angular=angular,
                orthonormalization=orthonormalization,
            )
            derivatives_fn = functools.partial(
                generic_derivatives_fn,
                radial_size=radial_basis.size,
                angular=angular,
                orthonormalization=orthonormalization,
            )

            if self.n_spline_points is not None:
                positions = np.linspace(0, self.cutoff.radius, self.n_spline_points)
                spline = Spline(0, self.cutoff.radius)
                spline.add_points(
                    positions=positions,
                    values=values_fn(positions),
                    derivatives=derivatives_fn(positions),
                )
            else:
                spline = Spline.with_accuracy(
                    start=0,
                    stop=self.cutoff.radius,
                    values_fn=values_fn,
                    derivatives_fn=derivatives_fn,
                    accuracy=(
                        1e-8
                        if self.basis.spline_accuracy is None
                        else self.basis.spline_accuracy
                    ),
                )

            by_angular[angular] = SplinedRadialBasis(
                spline=spline,
                max_radial=radial_basis.max_radial,
                radius=self.cutoff.radius,
            )

        hypers["basis"] = Explicit(by_angular=by_angular, spline_accuracy=None)
        return hypers

    def _radial_integral(
        self,
        radial: int,
        angular: int,
        distances: np.ndarray,
        derivative: bool,
    ) -> np.ndarray:
        """Compute the SOAP radial integral with a given density and radial basis"""
        if isinstance(self.density, DiracDelta):
            radial_basis = self.basis.radial_basis(angular)
            return radial_basis.compute_primitive(
                distances, radial, derivative=derivative
            )
        elif isinstance(self.density, Gaussian):
            if derivative:
                return self._ri_gaussian_density_derivative(radial, angular, distances)
            else:
                return self._ri_gaussian_density(radial, angular, distances)
        else:
            return self._ri_custom_density(
                radial, angular, distances, derivative=derivative
            )

    def _ri_gaussian_density(
        self,
        radial: int,
        angular: int,
        distances: np.ndarray,
    ) -> np.ndarray:
        # This code is derived in
        # https://metatensor.github.io/featomic/latest/devdoc/explanations/radial-integral.html,
        # and follows the same naming convention as the documentation.
        sigma_sq = self.density.width**2

        prefactor = (
            (4 * np.pi)
            / (np.pi * sigma_sq) ** (3 / 4)
            * np.exp(-0.5 * distances**2 / sigma_sq)
        )

        radial_basis = self.basis.radial_basis(angular)

        def integrand(r: float, n: int, ell: int, rij: np.array) -> np.ndarray:
            Rnl = radial_basis.compute_primitive(r, n, derivative=False)

            z = r * rij / sigma_sq
            bessel = _spherical_bessel_scaled(ell, z)

            return r**2 * Rnl * np.exp(z - 0.5 * r**2 / sigma_sq) * bessel

        integral = scipy.integrate.quad_vec(
            f=integrand,
            a=0,
            b=radial_basis.integration_radius,
            args=(radial, angular, distances),
        )[0]
        return prefactor * integral

    def _ri_gaussian_density_derivative(
        self,
        radial: int,
        angular: int,
        distances: np.ndarray,
    ) -> np.ndarray:
        # This code is derived in
        # https://metatensor.github.io/featomic/latest/devdoc/explanations/radial-integral.html,
        # and follows the same naming convention as the documentation.
        sigma_sq = self.density.width**2

        prefactor = (
            (4 * np.pi)
            / (np.pi * sigma_sq) ** (3 / 4)
            * np.exp(-0.5 * distances**2 / sigma_sq)
        )

        radial_basis = self.basis.radial_basis(angular)

        def integrand(r: float, n: int, ell: int, rij: np.array) -> np.ndarray:
            Rnl = radial_basis.compute_primitive(r, n, derivative=False)

            z = r * rij / sigma_sq
            one_over_z = np.divide(1.0, z, out=np.zeros_like(z), where=z != 0)

            # using recurrence relation from https://dlmf.nist.gov/10.51#E5 for the
            # derivative
            bessel_ell = _spherical_bessel_scaled(ell, z)
            bessel_ell_p1 = _spherical_bessel_scaled(ell + 1, z)
            bessel_grad = bessel_ell_p1 + ell * one_over_z * bessel_ell

            # The formula above is wrong for z=0, so let's replace it manually
            if ell == 1:
                bessel_grad = np.where(z == 0, 1.0 / 3.0, bessel_grad)

            return r**3 * Rnl * np.exp(z - 0.5 * r**2 / sigma_sq) * bessel_grad

        radial_integral = self._ri_gaussian_density(radial, angular, distances)
        grad_integral = scipy.integrate.quad_vec(
            f=integrand,
            a=0,
            b=radial_basis.integration_radius,
            args=(radial, angular, distances),
        )[0]
        return (prefactor * grad_integral - distances * radial_integral) / sigma_sq

    def _ri_custom_density(
        self,
        radial: int,
        angular: int,
        distances: np.ndarray,
        derivative: bool,
    ) -> np.ndarray:
        # This code is derived in
        # https://metatensor.github.io/featomic/latest/devdoc/explanations/radial-integral.html,
        # and follows the same naming convention as the documentation.

        P_ell = scipy.special.legendre(angular)
        radial_basis = self.basis.radial_basis(angular)

        if derivative:

            def integrand(u: float, r: float, n: int, ell: int, rij: float) -> float:
                arg = np.sqrt(r**2 + rij**2 - 2 * r * rij * u)

                Rnl = radial_basis.compute_primitive(r, n, derivative=False)
                density_grad = self.density.compute(arg, derivative=True)

                return r**2 * Rnl * P_ell(u) * (rij - u * r) * density_grad / arg

        else:

            def integrand(u: float, r: float, n: int, ell: int, rij: float) -> float:
                arg = np.sqrt(r**2 + rij**2 - 2 * r * rij * u)

                Rnl = radial_basis.compute_primitive(r, n, derivative=False)
                density = self.density.compute(arg, derivative=False)

                return r**2 * Rnl * P_ell(u) * density

        radial_integral = np.zeros(len(distances))

        for i, rij in enumerate(distances):
            radial_integral[i], _ = scipy.integrate.dblquad(
                func=integrand,
                # integration bounds for `r``
                a=0,
                b=radial_basis.integration_radius,
                # integration bounds for `u`
                gfun=-1,
                hfun=1,
                args=(radial, angular, rij),
            )

        return 2 * np.pi * radial_integral


class LodeSpliner:
    r"""Compute an explicit spline of the radial integral for LODE/k-space calculators.

    This allows a great deal of customization in the radial basis function used (any
    child class of :py:class:`RadialBasis`). This way, users can define custom basis,
    and use them with the LODE calculators. For more information about the radial
    integral, you can refer to :ref:`this document <radial-integral>`.

    This class should be used only in combination with k-space/LODE calculators like
    :class:`featomic.LodeSphericalExpansion`. For real space spherical expansions you
    should use :class:`SoapSpliner`.

    Example
    -------

    First let's define the hyper parameters for the LODE spherical expansions. It is
    important to note that only class-based hyper parameters are supported (in
    opposition to ``dict`` based hyper parameters).

    >>> import featomic

    >>> density = featomic.density.SmearedPowerLaw(smearing=1.0, exponent=1)
    >>> basis = featomic.basis.TensorProduct(
    ...     max_angular=4,
    ...     radial=featomic.basis.Gto(max_radial=5, radius=2),
    ... )

    From here we can initialize the spliner instance

    >>> spliner = LodeSpliner(density=density, basis=basis)

    You can then give the result of :meth:`LodeSpliner.get_hypers()` directly the
    calculator:

    >>> calculator = featomic.LodeSphericalExpansion(**spliner.get_hypers())

    .. seealso::
        :class:`SoapSpliner` for a spliner class that works with
        :class:`featomic.SphericalExpansion`
    """

    def __init__(
        self,
        density: AtomicDensity,
        basis: ExpansionBasis,
        k_cutoff: Optional[float] = None,
        n_spline_points: Optional[int] = None,
    ):
        """
        :param density: atomic density that should be expanded for all neighbors in the
            local atomic environment
        :param basis: basis function to use to expand the neighbors' atomic density.
            Currently only :py:class:`TensorProduct` expansion basis are supported.
        :param k_cutoff: Spherical cutoff in reciprocal space.
        :param n_spline_points: number of spline points to use. If ``None``, points will
            be added to the spline until the accuracy is at least the requested
            ``basis.spline_accuracy``.
        """
        if not HAS_SCIPY:
            raise ImportError("LodeSpliner class requires scipy")

        if not isinstance(density, SmearedPowerLaw):
            raise TypeError(
                "only `SmearedPowerLaw` `density` is supported by LODE, "
                f"got {type(density)}"
            )

        if not isinstance(basis, ExpansionBasis):
            raise TypeError(
                f"`basis` should be an `ExpansionBasis object, got {type(basis)}"
            )

        if k_cutoff is None:
            self.k_cutoff = 1.2 * np.pi / density.smearing
        else:
            self.k_cutoff = float(k_cutoff)

        self.density = density
        self.basis = basis

        self.n_spline_points = n_spline_points

    def get_hypers(self):
        """Get the LODE hyper-parameters for the splined basis and density."""
        # Contrary to the SOAP spliner, we don't transform the density to a delta
        # density, since the splines are applied directly to k-vectors (and not to the
        # density).
        hypers = {
            "k_cutoff": self.k_cutoff,
            "density": self.density,
        }

        def generic_values_fn(positions, radial_size, angular, orthonormalization):
            integrals = np.vstack(
                [
                    self._radial_integral(radial, angular, positions, derivative=False)
                    for radial in range(radial_size)
                ]
            )
            return (orthonormalization @ integrals).T

        def generic_derivatives_fn(positions, radial_size, angular, orthonormalization):
            integrals = np.vstack(
                [
                    self._radial_integral(radial, angular, positions, derivative=True)
                    for radial in range(radial_size)
                ]
            )
            return (orthonormalization @ integrals).T

        # We still transform whatever basis the user provided to an "explicit" basis,
        # since the radial integral will be different for different angular channels.
        by_angular = {}
        for angular in self.basis.angular_channels():
            radial_basis = self.basis.radial_basis(angular)
            orthonormalization = radial_basis._get_orthonormalization_matrix()

            values_fn = functools.partial(
                generic_values_fn,
                radial_size=radial_basis.size,
                angular=angular,
                orthonormalization=orthonormalization,
            )
            derivatives_fn = functools.partial(
                generic_derivatives_fn,
                radial_size=radial_basis.size,
                angular=angular,
                orthonormalization=orthonormalization,
            )

            if self.n_spline_points is not None:
                positions = np.linspace(0, self.k_cutoff, self.n_spline_points)
                spline = Spline(0, self.k_cutoff)
                spline.add_points(
                    positions=positions,
                    values=values_fn(positions),
                    derivatives=derivatives_fn(positions),
                )
            else:
                spline = Spline.with_accuracy(
                    start=0,
                    stop=self.k_cutoff,
                    values_fn=values_fn,
                    derivatives_fn=derivatives_fn,
                    accuracy=(
                        1e-8
                        if self.basis.spline_accuracy is None
                        else self.basis.spline_accuracy
                    ),
                )

            if angular == 0:
                center_contribution = self._center_contribution()
            else:
                center_contribution = None

            by_angular[angular] = SplinedRadialBasis(
                spline=spline,
                max_radial=radial_basis.max_radial,
                radius=self.k_cutoff,
                lode_center_contribution=center_contribution,
            )

        hypers["basis"] = Explicit(by_angular=by_angular, spline_accuracy=None)
        return hypers

    def _radial_integral(
        self,
        radial: int,
        angular: int,
        distances: np.ndarray,
        derivative: bool,
    ) -> np.ndarray:
        radial_basis = self.basis.radial_basis(angular)
        if derivative:

            def integrand(r: float, n: int, ell: int, rij: np.ndarray) -> np.ndarray:
                Rnl = radial_basis.compute_primitive(r, n, derivative=False)
                spherical_jn = scipy.special.spherical_jn(ell, r * rij, derivative=True)
                return r**3 * Rnl * spherical_jn

        else:

            def integrand(r: float, n: int, ell: int, rij: np.ndarray) -> np.ndarray:
                Rnl = radial_basis.compute_primitive(r, n, derivative=False)
                spherical_jn = scipy.special.spherical_jn(
                    ell, r * rij, derivative=False
                )
                return r**2 * Rnl * spherical_jn

        return scipy.integrate.quad_vec(
            f=integrand,
            a=0,
            b=radial_basis.integration_radius,
            args=(radial, angular, distances),
        )[0]

    def _center_contribution(self) -> np.ndarray:
        assert isinstance(self.density, SmearedPowerLaw)

        radial_basis = self.basis.radial_basis(angular=0)

        def integrand(r: float, n: int) -> np.ndarray:
            return (
                r**2
                * radial_basis.compute_primitive(r, n, derivative=False)
                * self.density.compute(r, derivative=False)
            )

        integrals = []
        for radial in range(radial_basis.size):
            integrals.append(
                scipy.integrate.quad(
                    func=integrand,
                    a=0,
                    b=radial_basis.integration_radius,
                    args=(radial,),
                )[0]
            )

        return np.sqrt(4 * np.pi) * np.array(integrals)



"""
This module provides a convenience calculator for computing a two-center equivariant
power spectrum, or equivariant power spectrum by pair
"""

import json
from typing import List, Optional, Union

from . import _dispatch
from ._backend import (
    CalculatorBase,
    Device,
    DType,
    IntoSystem,
    Labels,
    TensorMap,
    TorchModule,
    operations,
)
from ._cg_product import ClebschGordanProduct
from ._density_correlations import _filter_redundant_keys


class EquivariantPowerSpectrumByPair(TorchModule):
    r"""
    Computes a general equivariant power spectrum by pair descriptor of two calculators,
    the second being a calculator by pair.

    Example
    -------

    As an example we calculate the equivariant power spectrum by pair for a spherical
    expansion and a spherical expansion by pair for a NaCl crystal.

    >>> import featomic
    >>> import ase

    Construct the NaCl crystal

    >>> atoms = ase.Atoms(
    ...     symbols="NaCl",
    ...     positions=[[0, 0, 0], [0.5, 0.5, 0.5]],
    ...     pbc=True,
    ...     cell=[1, 1, 1],
    ... )

    Define the hyper parameters for the short-range spherical expansion

    >>> spex_hypers = {
    ...     "cutoff": {
    ...         "radius": 3.0,
    ...         "smoothing": {"type": "ShiftedCosine", "width": 0.5},
    ...     },
    ...     "density": {
    ...         "type": "Gaussian",
    ...         "width": 0.3,
    ...     },
    ...     "basis": {
    ...         "type": "TensorProduct",
    ...         "max_angular": 2,
    ...         "radial": {"type": "Gto", "max_radial": 5},
    ...     },
    ... }

    Define the hyper parameters for the spherical expansion by pair

    >>> spex_by_pair_hypers = {
    ...     "cutoff": {
    ...         "radius": 5.0,
    ...         "smoothing": {"type": "ShiftedCosine", "width": 0.5},
    ...     },
    ...     "density": {
    ...         "type": "Gaussian",
    ...         "width": 0.3,
    ...     },
    ...     "basis": {
    ...         "type": "TensorProduct",
    ...         "max_angular": 1,
    ...         "radial": {"type": "Gto", "max_radial": 5},
    ...     },
    ... }

    Construct the calculators

    >>> spex_calculator = featomic.SphericalExpansion(**spex_hypers)
    >>> spex_by_pair_calculator = featomic.SphericalExpansionByPair(
    ...     **spex_by_pair_hypers
    ... )

    Construct the power spectrum by pair calculators and compute the spherical expansion

    >>> calculator = featomic.clebsch_gordan.EquivariantPowerSpectrumByPair(
    ...     spex_calculator, spex_by_pair_calculator
    ... )
    >>> power_spectrum_by_pair = calculator.compute(atoms, neighbors_to_properties=True)

    The resulting equivariants are stored as :py:class:`metatensor.TensorMap` as for any
    other calculator. The keys contain the symmetry information:

    >>> power_spectrum_by_pair.keys
    Labels(
        o3_lambda  o3_sigma  first_atom_type  second_atom_type
            0         1            11                11
            1         1            11                11
            0         1            11                17
            1         1            11                17
            1         -1           11                11
            2         1            11                11
            1         -1           11                17
            2         1            11                17
            0         1            17                11
            1         1            17                11
            0         1            17                17
            1         1            17                17
            1         -1           17                11
            2         1            17                11
            1         -1           17                17
            2         1            17                17
    )

    The block properties contain the angular order of the combined blocks ("l_1",
    "l_2"), along with the neighbor type of the full spherical expansion and the radial
    channel indices.

    >>> power_spectrum_by_pair[0].properties.names
    ['l_1', 'l_2', 'neighbor_1_type', 'n_1', 'n_2']

    .. seealso::
        An equivariant power spectrum calculator for single-center descriptors can be
        found at :py:class:`featomic.clebsch_gordan.EquivariantPowerSpectrum`.
    """

    def __init__(
        self,
        calculator_1: CalculatorBase,
        calculator_2: CalculatorBase,
        neighbor_types: Optional[List[int]] = None,
        *,
        dtype: Optional[DType] = None,
        device: Optional[Device] = None,
    ):
        """

        Constructs the equivariant power spectrum by pair calculator.

        :param calculator_1: first calculator that computes a density descriptor, either
            a :py:class:`featomic.SphericalExpansion` or
            :py:class:`featomic.LodeSphericalExpansion`.
        :param calculator_2: second calculator that computes a density by pair
            descriptor, it must be :py:class:`featomic.SphericalExpansionByPair` for
            now.
        :param neighbor_types: List of ``"neighbor_type"`` to use in the properties of
            the output. This option might be useful when running the calculation on
            subset of a whole dataset and trying to join along the ``sample`` dimension
            after the calculation. If ``None``, blocks are filled with
            ``"neighbor_type"`` found in the systems. This parameter is only used if
            ``neighbors_to_properties=True`` is passed to the :py:meth:`compute` method.
        :param dtype: the scalar type to use to store coefficients
        :param device: the computational device to use for calculations.
        """

        super().__init__()
        self.calculator_1 = calculator_1
        self.calculator_2 = calculator_2
        self.neighbor_types = neighbor_types
        self.dtype = dtype
        self.device = device

        supported_calculators_1 = ["lode_spherical_expansion", "spherical_expansion"]
        supported_calculators_2 = ["spherical_expansion_by_pair"]

        if self.calculator_1.c_name not in supported_calculators_1:
            raise ValueError(
                f"Only [{', '.join(supported_calculators_1)}] are supported for "
                f"`calculator_1`, got '{self.calculator_1.c_name}'"
            )

        parameters_1 = json.loads(calculator_1.parameters)

        # For the moment, only spherical expansion by pair is supported for calculator_2
        if self.calculator_2.c_name not in supported_calculators_2:
            raise ValueError(
                f"Only [{', '.join(supported_calculators_2)}] are supported for "
                f"`calculator_2`, got '{self.calculator_2.c_name}'"
            )

        parameters_2 = json.loads(calculator_2.parameters)
        if parameters_1["basis"]["type"] != "TensorProduct":
            raise ValueError("only 'TensorProduct' basis is supported for calculator_1")

        if parameters_2["basis"]["type"] != "TensorProduct":
            raise ValueError("only 'TensorProduct' basis is supported for calculator_2")

        self._cg_product = ClebschGordanProduct(
            max_angular=parameters_1["basis"]["max_angular"]
            + parameters_2["basis"]["max_angular"],
            cg_backend=None,
            keys_filter=_filter_redundant_keys,
            arrays_backend=None,
            dtype=dtype,
            device=device,
        )

    @property
    def name(self):
        """Name of this calculator."""
        return "EquivariantPowerSpectrumByPair"

    def compute(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        selected_keys: Optional[Labels] = None,
        selected_samples: Optional[Labels] = None,
        neighbors_to_properties: bool = False,
    ) -> TensorMap:
        """
        Computes an equivariant power spectrum by pair, that can be thought of as a
        "Lambda-SOAP by pair" when doing a correlation of the SOAP density with a
        spherical expansion by pair.

        First computes a :py:class:`SphericalExpansion` density descriptor of body order
        2. Then a :py:class:`SphericalExpansionByPair` two-center density descriptor of
        body order 2 is computed.

        Before performing the Clebsch-Gordan tensor product, the spherical expansion
        density can be densified by moving the key dimension "neighbor_type" to the
        block properties. This is controlled by the ``neighbors_to_properties``
        parameter. Depending on the specific systems descriptors are being computed for,
        the sparsity of the spherical expansion can affect the computational cost of the
        Clebsch-Gordan tensor product.

        If ``neighbors_to_properties=True`` and ``neighbor_types`` have been passed to
        the constructor, property dimensions are created for all of these global atom
        types when moving the key dimension to properties. This ensures that the output
        properties dimension is of consistent size across all systems passed in
        ``systems``.

        Finally a single Clebsch-Gordan tensor product is taken to produce a body order
        3 equivariant power spectrum by pair.

        :param selected_keys: :py:class:`Labels`, the output keys to computed. If
            ``None``, all keys are computed. Subsets of key dimensions can be passed to
            compute output blocks that match in these dimensions.
        :param selected_samples: :py:class:`Labels`, Set of samples on which to run the
            calculation. Use ``None`` to run the calculation on all samples in
            the systems (this is the default). Gets passed to ``calculator_1``  and
            ``calculator_2``, therefore requiring that both calculators support sample
            selection.
        :param neighbors_to_properties: :py:class:`bool`, if true, densifies the
            spherical expansion by moving key dimension "neighbor_type" to properties
            prior to performing the Clebsch Gordan product step. Defaults to false.

        :return: :py:class:`TensorMap`, the output equivariant power spectrum by pair.
        """
        return self._equivariant_power_spectrum_by_pair(
            systems=systems,
            selected_keys=selected_keys,
            selected_samples=selected_samples,
            neighbors_to_properties=neighbors_to_properties,
            compute_metadata=False,
        )

    def forward(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        selected_keys: Optional[Labels] = None,
        selected_samples: Optional[Labels] = None,
        neighbors_to_properties: bool = False,
    ) -> TensorMap:
        """
        Calls the :py:meth:`compute` method.

        This is intended for :py:class:`torch.nn.Module` compatibility, and should be
        ignored in pure Python mode.

        See :py:meth:`compute` for a full description of the parameters.
        """
        return self.compute(
            systems=systems,
            selected_keys=selected_keys,
            selected_samples=selected_samples,
            neighbors_to_properties=neighbors_to_properties,
        )

    def compute_metadata(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        selected_keys: Optional[Labels] = None,
        selected_samples: Optional[Labels] = None,
        neighbors_to_properties: bool = False,
    ) -> TensorMap:
        """
        Returns the metadata-only :py:class:`TensorMap` that would be output by the
        function :py:meth:`compute` for the same calculator under the same settings,
        without performing the actual Clebsch-Gordan tensor products in the second step.

        See :py:meth:`compute` for a full description of the parameters.
        """
        return self._equivariant_power_spectrum_by_pair(
            systems=systems,
            selected_keys=selected_keys,
            selected_samples=selected_samples,
            neighbors_to_properties=neighbors_to_properties,
            compute_metadata=True,
        )

    def _equivariant_power_spectrum_by_pair(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        selected_keys: Optional[Labels],
        selected_samples: Optional[Labels],
        neighbors_to_properties: bool,
        compute_metadata: bool,
    ) -> TensorMap:
        """
        Computes the equivariant power spectrum by pair, either fully or just metadata
        """
        # Compute density
        density_1 = self.calculator_1.compute(
            systems, selected_samples=selected_samples
        )
        # Rename "center_type" dimension to match "first_atom_type"
        density_1 = operations.rename_dimension(
            density_1, "keys", "center_type", "first_atom_type"
        )
        # Rename "neighbor_type" dimension so they are correlated
        density_1 = operations.rename_dimension(
            density_1, "keys", "neighbor_type", "neighbor_1_type"
        )
        # Rename samples to match the pair
        density_1 = operations.rename_dimension(
            density_1, "samples", "atom", "first_atom"
        )
        # Rename properties so they are correlated
        density_1 = operations.rename_dimension(density_1, "properties", "n", "n_1")

        # Compute pair density
        if selected_samples is not None:
            if "atom" in selected_samples.names:
                new_names = selected_samples.names
                new_names[selected_samples.names.index("atom")] = "first_atom"
                selected_samples = Labels(new_names, selected_samples.values)
        density_2 = self.calculator_2.compute(
            systems, selected_samples=selected_samples
        )

        # Rename properties so they are correlated
        density_2 = operations.rename_dimension(density_2, "properties", "n", "n_2")

        if neighbors_to_properties:
            if self.neighbor_types is None:  # just move neighbor type
                keys_to_move_1 = "neighbor_1_type"
            else:  # use the user-specified types
                values = _dispatch.list_to_array(
                    array=density_1.keys.values,
                    data=[[t] for t in self.neighbor_types],
                )
                keys_to_move_1 = Labels(names="neighbor_1_type", values=values)

            density_1 = density_1.keys_to_properties(keys_to_move_1)

        # Compute the power spectrum
        if compute_metadata:
            pow_spec = self._cg_product.compute_metadata(
                tensor_1=density_1,
                tensor_2=density_2,
                o3_lambda_1_new_name="l_1",
                o3_lambda_2_new_name="l_2",
                selected_keys=selected_keys,
            )
        else:
            pow_spec = self._cg_product.compute(
                tensor_1=density_1,
                tensor_2=density_2,
                o3_lambda_1_new_name="l_1",
                o3_lambda_2_new_name="l_2",
                selected_keys=selected_keys,
            )

        # Move the CG combination info keys to properties
        pow_spec = pow_spec.keys_to_properties(["l_1", "l_2"])

        return pow_spec



import copy
import warnings

import numpy as np


try:
    import ase
    from ase import neighborlist

    HAVE_ASE = True
except ImportError:
    HAVE_ASE = False

from .base import SystemBase


class AseSystem(SystemBase):
    """Implements :py:class:`featomic.SystemBase` using `ase.Atoms`_.

    Gets the data and `ase.neighborlist.neighbor_list`_ to
    compute the neighbor list.

    .. _ase.Atoms: https://wiki.fysik.dtu.dk/ase/ase/atoms.html

    .. _ase.neighborlist.neighbor_list:
        https://wiki.fysik.dtu.dk/ase/ase/neighborlist.html#ase.neighborlist.neighbor_list
    """

    @staticmethod
    def can_wrap(o):
        return isinstance(o, ase.Atoms)

    def __init__(self, atoms):
        """:param atoms: `ase.Atoms`_ object to be wrapped in this ``AseSystem``"""
        super().__init__()
        if not isinstance(atoms, ase.Atoms):
            raise TypeError("this class expects ASE.Atoms objects")

        # normalize pbc to three values
        if isinstance(atoms.pbc, bool):
            atoms_pbc = [atoms.pbc, atoms.pbc, atoms.pbc]
        else:
            atoms_pbc = copy.deepcopy(atoms.pbc)

        self._cell = np.array(atoms.cell)
        # validate pcb consistency with the cell matrix
        if np.all(np.abs(self._cell) < 1e-9):
            if np.any(atoms_pbc):
                raise ValueError(
                    "periodic boundary conditions are enabled, "
                    "but the cell matrix is zero everywhere. "
                    "You should set pbc to `False`, or the cell to its value."
                )
        elif np.any(np.bitwise_not(atoms_pbc)):
            if np.all(np.bitwise_not(atoms_pbc)):
                warnings.warn(
                    "periodic boundary conditions are disabled, but the cell "
                    "matrix is not zero, we will set the cell to zero.",
                    stacklevel=1,
                )
                self._cell[:, :] = 0.0
            else:
                raise ValueError(
                    "different periodic boundary conditions on different axis "
                    "are not supported"
                )

        self._atoms = atoms
        self._pairs = []
        self._pairs_by_atom = []
        self._last_cutoff = None

    def size(self):
        return len(self._atoms)

    def types(self):
        return self._atoms.numbers

    def positions(self):
        return self._atoms.positions

    def cell(self):
        return self._cell

    def compute_neighbors(self, cutoff):
        if self._last_cutoff == cutoff:
            return

        self._pairs = []

        nl_result = neighborlist.neighbor_list("ijdDS", self._atoms, cutoff)
        for i, j, d, D, S in zip(*nl_result):
            # we want a half neighbor list, so drop all duplicated neighbors
            if j < i:
                continue
            elif i == j:
                if S[0] == 0 and S[1] == 0 and S[2] == 0:
                    # only create pairs with the same atom twice if the pair spans more
                    # than one unit cell
                    continue
                elif S[0] + S[1] + S[2] < 0 or (
                    (S[0] + S[1] + S[2] == 0) and (S[2] < 0 or (S[2] == 0 and S[1] < 0))
                ):
                    # When creating pairs between an atom and one of its periodic
                    # images, the code generate multiple redundant pairs (e.g. with
                    # shifts 0 1 1 and 0 -1 -1); and we want to only keep one of these.
                    # We keep the pair in the positive half plane of shifts.
                    continue

            self._pairs.append((i, j, d, D, S))

        self._pairs_by_atom = []
        for _ in range(self.size()):
            self._pairs_by_atom.append([])

        for i, j, d, D, S in self._pairs:
            self._pairs_by_atom[i].append((i, j, d, D, S))
            self._pairs_by_atom[j].append((i, j, d, D, S))

    def pairs(self):
        return self._pairs

    def pairs_containing(self, atom):
        return self._pairs_by_atom[atom]



from ._c_api import FEATOMIC_SUCCESS
from ._c_lib import _get_library


class FeatomicError(Exception):
    """Exceptions thrown for all errors in featomic."""

    def __init__(self, message, status=None):
        super(Exception, self).__init__(message)

        self.message = message
        """``str``, error message for this exception"""

        self.status = status
        """``Optional[int]``, status code for this exception"""


LAST_EXCEPTION = None


def _save_exception(e):
    global LAST_EXCEPTION
    LAST_EXCEPTION = e


def _check_featomic_status_t(status):
    if status == FEATOMIC_SUCCESS:
        return
    elif status > FEATOMIC_SUCCESS:
        raise FeatomicError(last_error(), status)
    elif status < FEATOMIC_SUCCESS:
        global LAST_EXCEPTION
        e = LAST_EXCEPTION
        LAST_EXCEPTION = None
        raise FeatomicError(last_error(), status) from e


def _check_featomic_pointer(pointer):
    if not pointer:
        raise FeatomicError(last_error())


def last_error():
    """Get the last error message on this thread."""
    lib = _get_library()
    message = lib.featomic_last_error()
    return message.decode("utf8")



import ctypes
from ctypes import POINTER, c_double, c_void_p, pointer

import numpy as np

from .._c_api import c_uintptr_t, featomic_pair_t, featomic_system_t
from ..status import _save_exception


def catch_exceptions(function):
    """Decorate a function catching any exception."""

    def inner(*args, **kwargs):
        try:
            function(*args, **kwargs)
        except Exception as e:
            _save_exception(e)
            return -1
        return 0

    return inner


class SystemBase:
    """Base class implementing the ``System`` trait in featomic.

    Developers should implement this class to add new kinds of system that work with
    featomic.

    Most users should use one of the already provided implementation, such as
    :py:class:`featomic.systems.AseSystem` or
    :py:class:`featomic.systems.ChemfilesSystem` instead of using this class directly.

    A very simple implementation of the interface is given below as starting point. The
    example does not implement a neighbor list, and it can only be used by setting
    ``use_native_system=True`` in
    :py:meth:`featomic.calculators.CalculatorBase.compute`, to transfer the data to the
    native code and compute the neighbors list there.

    >>> import featomic
    >>> import numpy as np
    >>>
    >>> class SimpleSystem(featomic.systems.SystemBase):
    ...     def __init__(self, types, positions, cell):
    ...         super().__init__()
    ...
    ...         # types should be a 1D array of integers
    ...         types = np.asarray(types)
    ...         assert len(types.shape) == 1
    ...         assert types.dtype == np.int32
    ...         self._types = types
    ...
    ...         # positions should be a 2D array of float
    ...         positions = np.asarray(positions)
    ...         assert len(positions.shape) == 2
    ...         assert positions.shape[0] == types.shape[0]
    ...         assert positions.shape[1] == 3
    ...         assert positions.dtype == np.float64
    ...         self._positions = positions
    ...
    ...         # cell should be a 3x3 array of float
    ...         cell = np.asarray(cell)
    ...         assert len(cell.shape) == 2
    ...         assert cell.shape[0] == 3
    ...         assert cell.shape[1] == 3
    ...         assert cell.dtype == np.float64
    ...         self._cell = cell
    ...
    ...     def size(self):
    ...         return len(self._types)
    ...
    ...     def types(self):
    ...         return self._types
    ...
    ...     def positions(self):
    ...         return self._positions
    ...
    ...     def cell(self):
    ...         return self._cell
    ...
    ...     def compute_neighbors(self, cutoff):
    ...         raise NotImplementedError("this system does not have a neighbors list")
    ...
    ...     def pairs(self):
    ...         raise NotImplementedError("this system does not have a neighbors list")
    ...
    ...     def pairs_containing(self, atom):
    ...         raise NotImplementedError("this system does not have a neighbors list")
    >>> system = SimpleSystem(
    ...     types=np.random.randint(2, size=25, dtype=np.int32),
    ...     positions=6 * np.random.uniform(size=(25, 3)),
    ...     cell=6 * np.eye(3),
    ... )
    >>>
    >>> calculator = featomic.SortedDistances(
    ...     cutoff=3.3,
    ...     max_neighbors=4,
    ...     separate_neighbor_types=True,
    ... )
    >>>
    >>> # this works, and uses our new system
    >>> calculator.compute(system)
    TensorMap with 4 blocks
    keys: center_type  neighbor_type
               0             0
               0             1
               1             0
               1             1
    >>> # this does not work, since the code is trying to get a neighbors list
    >>> try:
    ...     calculator.compute(system, use_native_system=False)
    ... except featomic.FeatomicError as e:
    ...     raise e.__cause__
    Traceback (most recent call last):
        ...
    NotImplementedError: this system does not have a neighbors list
    """

    def __init__(self):
        # keep reference to some data to prevent garbage collection while Rust
        # might be using the data
        self._keepalive = {}

    def _as_featomic_system_t(self):
        """Convert a child instance of :py:class:`SystemBase`.

        Instances are converted to a C compatible ``featomic_system_t``.
        """
        struct = featomic_system_t()
        self._keepalive["c_struct"] = struct

        # user_data is a pointer to the PyObject `self`
        struct.user_data = ctypes.cast(pointer(ctypes.py_object(self)), c_void_p)

        def get_self(ptr):
            """Extract ``self`` from a pointer to the PyObject."""
            self = ctypes.cast(ptr, POINTER(ctypes.py_object)).contents.value
            assert isinstance(self, SystemBase)
            return self

        @catch_exceptions
        def featomic_system_size(user_data, size):
            """
            Implementation of ``featomic_system_t::size`` using
            :py:func:`SystemBase.size`.
            """
            size[0] = c_uintptr_t(get_self(user_data).size())

        # use struct.XXX.__class__ to get the right type for all functions
        struct.size = struct.size.__class__(featomic_system_size)

        @catch_exceptions
        def featomic_system_types(user_data, data):
            """
            Implementation of ``featomic_system_t::types`` using
            :py:func:`SystemBase.types`.
            """
            self = get_self(user_data)

            types = np.asarray(self.types(), order="C", dtype=np.int32)
            data[0] = types.ctypes.data
            self._keepalive["types"] = types

        struct.types = struct.types.__class__(featomic_system_types)

        @catch_exceptions
        def featomic_system_positions(user_data, data):
            """
            Implementation of ``featomic_system_t::positions`` using
            :py:func:`SystemBase.positions`.
            """
            self = get_self(user_data)
            positions = np.asarray(self.positions(), order="C", dtype=c_double)

            assert len(positions.shape) == 2
            assert positions.shape[1] == 3

            data[0] = positions.ctypes.data
            self._keepalive["positions"] = positions

        struct.positions = struct.positions.__class__(featomic_system_positions)

        @catch_exceptions
        def featomic_system_cell(user_data, data):
            """
            Implementation of ``featomic_system_t::cell`` using
            :py:func:`SystemBase.cell`.
            """
            self = get_self(user_data)
            cell = np.asarray(self.cell(), order="C", dtype=c_double)
            assert cell.shape == (3, 3)

            data[0] = cell[0][0]
            data[1] = cell[0][1]
            data[2] = cell[0][2]
            data[3] = cell[1][0]
            data[4] = cell[1][1]
            data[5] = cell[1][2]
            data[6] = cell[2][0]
            data[7] = cell[2][1]
            data[8] = cell[2][2]

        struct.cell = struct.cell.__class__(featomic_system_cell)

        @catch_exceptions
        def featomic_system_compute_neighbors(user_data, cutoff):
            """
            Implementation of ``featomic_system_t::compute_neighbors`` using
            :py:func:`SystemBase.compute_neighbors`.
            """
            self = get_self(user_data)
            self.compute_neighbors(cutoff)

        struct.compute_neighbors = struct.compute_neighbors.__class__(
            featomic_system_compute_neighbors
        )

        @catch_exceptions
        def featomic_system_pairs(user_data, data, count):
            """
            Implementation of ``featomic_system_t::pairs`` using
            :py:func:`SystemBase.pairs`.
            """
            self = get_self(user_data)

            pairs = np.asarray(
                self.pairs(),
                order="C",
                dtype=featomic_pair_t,
            )

            count[0] = c_uintptr_t(len(pairs))
            data[0] = pairs.ctypes.data
            self._keepalive["pairs"] = pairs

        struct.pairs = struct.pairs.__class__(featomic_system_pairs)

        @catch_exceptions
        def featomic_system_pairs_containing(user_data, atom, data, count):
            """
            Implementation of ``featomic_system_t::pairs_containing`` using
            :py:func:`SystemBase.pairs_containing`.
            """
            self = get_self(user_data)

            pairs = np.asarray(
                self.pairs_containing(atom),
                order="C",
                dtype=featomic_pair_t,
            )

            count[0] = c_uintptr_t(len(pairs))
            data[0] = pairs.ctypes.data
            self._keepalive["pairs_containing"] = pairs

        struct.pairs_containing = struct.pairs_containing.__class__(
            featomic_system_pairs_containing
        )

        return struct

    def size(self):
        """Get the number of atoms in this system as an integer."""

        raise NotImplementedError("System.size method is not implemented")

    def types(self):
        """Get the atomic types of all atoms in the system.

        Get a list of integers or a 1D numpy array of integers (ideally 32-bit
        integers, otherwise they will be converted on the fly) containing the
        atomic types for each atom in the system. Different atomic types
        should be identified with a different value. These values are usually
        the element number, but don't have to be.
        """

        raise NotImplementedError("System.types method is not implemented")

    def positions(self):
        """Get the cartesian position of all atoms in this system.

        The returned positions must be convertible to a numpy array of
        shape ``(self.size(), 3)``, with a dtype of `np.float64`.
        """

        raise NotImplementedError("System.positions method is not implemented")

    def cell(self):
        """Get the 3x3 matrix representing unit cell of the system.

        The cell should be written in row major order, i.e. `[[ax, ay, az], [bx,
        by, bz], [cx, cy, cz]]`, where a/b/c are the unit cell vectors.

        If the cell is returned as a numpy array, it should have a dtype of
        `np.float64`.
        """

        raise NotImplementedError("System.cell method is not implemented")

    def compute_neighbors(self, cutoff):
        """Compute the neighbor list with the given ``cutoff``.

        Store it for later access using :py:func:`featomic.SystemBase.pairs` or
        :py:func:`featomic.SystemBase.pairs_containing`.
        """

        raise NotImplementedError("System.compute_neighbors method is not implemented")

    def pairs(self):
        """Atoms pairs in this system.

        The pairs are those which were computed by the last call
        :py:func:`SystemBase.compute_neighbors`

        Get all neighbor pairs in this system as a list of tuples ``(int, int, float,
        (float, float, float), (int, int, int))`` containing the indexes of the first
        and second atom in the pair, the distance between the atoms, the vector between
        them, and the cell shift. The vector should be ``position[first] -
        position[second] * + H * cell_shift`` where ``H`` is the cell matrix.
        Alternatively, this function can return a 1D numpy array with
        ``dtype=featomic_pair_t``.

        The list of pair should only contain each pair once (and not twice as ``i-j``
        and ``j-i``), should not contain self pairs (``i-i``); and should only contains
        pairs where the distance between atoms is actually bellow the cutoff passed in
        the last call to :py:func:`featomic.SystemBase.compute_neighbors`.

        This function is only valid to call after a call to
        :py:func:`featomic.SystemBase.compute_neighbors` to set the cutoff.
        """

        raise NotImplementedError("System.pairs method is not implemented")

    def pairs_containing(self, atom):
        """
        Get all neighbor pairs in this system containing the atom with index
        ``atom``.

        The return type of this function should be the same as
        :py:func:`featomic.SystemBase.pairs`. The same restrictions on the list
        of pairs also applies, with the additional condition that the pair
        ``i-j`` should be included both in the list returned by
        ``pairs_containing(i)`` and ``pairs_containing(j)``.
        """

        raise NotImplementedError("System.pairs_containing method is not implemented")



import abc
from typing import Optional

import numpy as np


class SmoothingFunction(metaclass=abc.ABCMeta):
    """
    Base class representing radial cutoff smoothing functions.

    You can inherit from this class to define new smoothing functions, implementing
    :py:meth:`compute` accordingly. If the new smoothing function has corresponding
    hyper parameters in the native calculators, you should also implement
    :py:meth:`get_hypers`.
    """

    def _featomic_hypers(self):
        return self.get_hypers()

    def get_hypers(self):
        raise NotImplementedError(
            f"this smoothing function ({self.__class__.__name__}) does not have "
            "matching hyper parameters in the native calculators"
        )

    @abc.abstractmethod
    def compute(
        self, cutoff: float, positions: np.ndarray, *, derivative: bool
    ) -> np.ndarray:
        """
        Compute the smoothing function on grid points at the given ``positions``.

        :param cutoff: spherical cutoff radius
        :param positions: positions of the grid points where the smoothing function
            should be evaluated
        :param derivative: should this function return the values of the smoothing
            function or it's derivatives
        """


class ShiftedCosine(SmoothingFunction):
    r"""
    Shifted cosine smoothing function, with the following form:

    .. math::

        f(r) = \begin{cases}
            1 & \text{for } r \le r_c - \sigma \\
            1/2 \left(1 + \frac{\cos(\pi (r - r_c + \sigma)}{\sigma} \right)
            & \text{for } r_c - \sigma \le r \le r_c \\
            0 & \text{for } r \gt r_c \\
        \end{cases}

    with :math:`r_c` the cutoff radius and :math:`\sigma` is the width of the smoothing
    (roughly how far from the cutoff smoothing should happen).

    :param width: width of the smoothing (:math:`\sigma` in the equation above)
    """

    def __init__(self, *, width: float):
        self.width = float(width)
        assert self.width >= 0

    def get_hypers(self):
        return {"type": "ShiftedCosine", "width": self.width}

    def compute(
        self, cutoff: float, positions: np.ndarray, *, derivative: bool
    ) -> np.ndarray:
        assert cutoff > self.width
        result = np.zeros_like(positions)

        mask = np.logical_and(positions >= cutoff - self.width, positions < cutoff)
        s = np.pi * (positions[mask] - cutoff + self.width) / self.width

        if derivative:
            result[mask] = -0.5 * np.pi * np.sin(s) / self.width
        else:
            result[positions < cutoff - self.width] = 1.0
            result[mask] = 0.5 * (1 + np.cos(s))

        return result


class Step(SmoothingFunction):
    """
    Step smoothing function, i.e. no smoothing.

    This function is equal to 1 inside the cutoff radius and to 0 outside of the cutoff
    radius, with a discontinuity at the cutoff.
    """

    def get_hypers(self):
        return {"type": "Step"}

    def compute(
        self, cutoff: float, positions: np.ndarray, *, derivative: bool
    ) -> np.ndarray:
        if derivative:
            return np.zeros_like(positions)
        else:
            return np.where(positions <= cutoff, 1.0, 0.0)


class Cutoff:
    """
    The ``Cutoff`` class contains the definition of local environments, where an atom
    environment is defined by all its neighbors inside a sphere centered on the atom
    with the given spherical cutoff ``radius``.

    During an atomistic simulation, atoms entering and exiting the sphere will create
    discontinuities. To prevent them, one can use a ``smoothing`` function, smoothing
    introducing new atoms inside the neighborhood.
    """

    def __init__(self, radius: float, smoothing: Optional[SmoothingFunction]):
        self.radius = float(radius)
        assert self.radius >= 0.0

        if smoothing is None:
            self.smoothing = Step()
        else:
            self.smoothing = smoothing

        assert isinstance(self.smoothing, SmoothingFunction)

    def _featomic_hypers(self):
        return {
            "radius": self.radius,
            "smoothing": self.smoothing,
        }



# fmt: off
# flake8: noqa
'''Automatically-generated file, do not edit!!!'''

import ctypes
import enum
import platform
from ctypes import CFUNCTYPE, POINTER

from metatensor._c_api import mts_labels_t, mts_tensormap_t
from numpy.ctypeslib import ndpointer


arch = platform.architecture()[0]
if arch == "32bit":
    c_uintptr_t = ctypes.c_uint32
elif arch == "64bit":
    c_uintptr_t = ctypes.c_uint64

FEATOMIC_SUCCESS = 0
FEATOMIC_INVALID_PARAMETER_ERROR = 1
FEATOMIC_JSON_ERROR = 2
FEATOMIC_UTF8_ERROR = 3
FEATOMIC_SYSTEM_ERROR = 128
FEATOMIC_BUFFER_SIZE_ERROR = 254
FEATOMIC_INTERNAL_ERROR = 255
FEATOMIC_LOG_LEVEL_ERROR = 1
FEATOMIC_LOG_LEVEL_WARN = 2
FEATOMIC_LOG_LEVEL_INFO = 3
FEATOMIC_LOG_LEVEL_DEBUG = 4
FEATOMIC_LOG_LEVEL_TRACE = 5


featomic_status_t = ctypes.c_int32
featomic_logging_callback_t = CFUNCTYPE(None, ctypes.c_int32, ctypes.c_char_p)


class featomic_calculator_t(ctypes.Structure):
    pass


class featomic_pair_t(ctypes.Structure):
    _fields_ = [
        ("first", c_uintptr_t),
        ("second", c_uintptr_t),
        ("distance", ctypes.c_double),
        ("vector", ctypes.c_double * 3),
        ("cell_shift_indices", ctypes.c_int32 * 3),
    ]


class featomic_system_t(ctypes.Structure):
    _fields_ = [
        ("user_data", ctypes.c_void_p),
        ("size", CFUNCTYPE(featomic_status_t, ctypes.c_void_p, POINTER(c_uintptr_t))),
        ("types", CFUNCTYPE(featomic_status_t, ctypes.c_void_p, POINTER(ndpointer(ctypes.c_int32, flags='C_CONTIGUOUS')))),
        ("positions", CFUNCTYPE(featomic_status_t, ctypes.c_void_p, POINTER(ndpointer(ctypes.c_double, flags='C_CONTIGUOUS')))),
        ("cell", CFUNCTYPE(featomic_status_t, ctypes.c_void_p, POINTER(ctypes.c_double))),
        ("compute_neighbors", CFUNCTYPE(featomic_status_t, ctypes.c_void_p, ctypes.c_double)),
        ("pairs", CFUNCTYPE(featomic_status_t, ctypes.c_void_p, POINTER(ndpointer(featomic_pair_t, flags='C_CONTIGUOUS')), POINTER(c_uintptr_t))),
        ("pairs_containing", CFUNCTYPE(featomic_status_t, ctypes.c_void_p, c_uintptr_t, POINTER(ndpointer(featomic_pair_t, flags='C_CONTIGUOUS')), POINTER(c_uintptr_t))),
    ]


class featomic_labels_selection_t(ctypes.Structure):
    _fields_ = [
        ("subset", POINTER(mts_labels_t)),
        ("predefined", POINTER(mts_tensormap_t)),
    ]


class featomic_calculation_options_t(ctypes.Structure):
    _fields_ = [
        ("gradients", POINTER(ctypes.c_char_p)),
        ("gradients_count", c_uintptr_t),
        ("use_native_system", ctypes.c_bool),
        ("selected_samples", featomic_labels_selection_t),
        ("selected_properties", featomic_labels_selection_t),
        ("selected_keys", POINTER(mts_labels_t)),
    ]


def setup_functions(lib):
    from .status import _check_featomic_status_t

    lib.featomic_last_error.argtypes = [
        
    ]
    lib.featomic_last_error.restype = ctypes.c_char_p

    lib.featomic_set_logging_callback.argtypes = [
        featomic_logging_callback_t
    ]
    lib.featomic_set_logging_callback.restype = _check_featomic_status_t

    lib.featomic_calculator.argtypes = [
        ctypes.c_char_p,
        ctypes.c_char_p
    ]
    lib.featomic_calculator.restype = POINTER(featomic_calculator_t)

    lib.featomic_calculator_free.argtypes = [
        POINTER(featomic_calculator_t)
    ]
    lib.featomic_calculator_free.restype = _check_featomic_status_t

    lib.featomic_calculator_name.argtypes = [
        POINTER(featomic_calculator_t),
        ctypes.c_char_p,
        c_uintptr_t
    ]
    lib.featomic_calculator_name.restype = _check_featomic_status_t

    lib.featomic_calculator_parameters.argtypes = [
        POINTER(featomic_calculator_t),
        ctypes.c_char_p,
        c_uintptr_t
    ]
    lib.featomic_calculator_parameters.restype = _check_featomic_status_t

    lib.featomic_calculator_cutoffs.argtypes = [
        POINTER(featomic_calculator_t),
        POINTER(POINTER(ctypes.c_double)),
        POINTER(c_uintptr_t)
    ]
    lib.featomic_calculator_cutoffs.restype = _check_featomic_status_t

    lib.featomic_calculator_compute.argtypes = [
        POINTER(featomic_calculator_t),
        POINTER(POINTER(mts_tensormap_t)),
        POINTER(featomic_system_t),
        c_uintptr_t,
        featomic_calculation_options_t
    ]
    lib.featomic_calculator_compute.restype = _check_featomic_status_t

    lib.featomic_profiling_clear.argtypes = [
        
    ]
    lib.featomic_profiling_clear.restype = _check_featomic_status_t

    lib.featomic_profiling_enable.argtypes = [
        ctypes.c_bool
    ]
    lib.featomic_profiling_enable.restype = _check_featomic_status_t

    lib.featomic_profiling_get.argtypes = [
        ctypes.c_char_p,
        ctypes.c_char_p,
        c_uintptr_t
    ]
    lib.featomic_profiling_get.restype = _check_featomic_status_t



import numpy as np

from .base import SystemBase


try:
    import pyscf

    HAVE_PYSCF = True
except ImportError:
    HAVE_PYSCF = False


# copied from pyscf/data/elements.py
def _std_symbol_without_ghost(symb_or_chg):
    """For a given atom symbol (lower case or upper case) or charge, return the
    standardized atom symbol
    """
    if isinstance(symb_or_chg, str):
        symb_or_chg = str(symb_or_chg.upper())
        rawsymb = pyscf.data.elements._rm_digit(symb_or_chg)
        if rawsymb in pyscf.data.elements._ELEMENTS_UPPER:
            return pyscf.data.elements._ELEMENTS_UPPER[rawsymb]
        elif len(rawsymb) > 1 and (symb_or_chg[0] == "X" and symb_or_chg[:2] != "XE"):
            rawsymb = rawsymb[1:]  # Remove the prefix X
            return pyscf.data.elements._ELEMENTS_UPPER[rawsymb]
        elif len(rawsymb) > 5 and rawsymb[:5] == "GHOST":
            rawsymb = rawsymb[5:]  # Remove the prefix GHOST
            return pyscf.data.elements._ELEMENTS_UPPER[rawsymb]
        else:
            raise RuntimeError("Unsupported atom symbol %s" % symb_or_chg)
    else:
        return pyscf.data.elements.ELEMENTS[symb_or_chg]


class PyscfSystem(SystemBase):
    """Implements :py:class:`featomic.SystemBase` wrapping a
    `pyscf.gto.mole.Mole`_ or `pyscf.pbc.gto.cell.Cell`_.

    Since pyscf does not offer a neighbors list, this implementation of system can only
    be used with ``use_native_system=True`` in
    :py:func:`featomic.calculators.CalculatorBase.compute`.

    Atomic type are assigned as the atomic number if the atom ``type`` is one of the
    periodic table elements; or their opposite if they are ghost atoms. (Pyscf does not
    seem to support anything else)

    Please note that while pyscf uses Bohrs as length units internally, we convert those
    back into Angströms for featomic. A pyscf object's "unit" attribute determines the
    units of the coordinates given *to pyscf*, which are by default angströms.

    .. _pyscf.gto.mole.Mole: https://pyscf.org/user/gto.html
    .. _pyscf.pbc.gto.cell.Cell: https://pyscf.org/user/pbc/gto.html
    """

    @staticmethod
    def can_wrap(o):
        # assumption: if we have a periodic system, then pyscf.pbc is defined
        if hasattr(pyscf, "pbc"):
            return isinstance(o, (pyscf.gto.mole.Mole, pyscf.pbc.gto.cell.Cell))
        else:
            return isinstance(o, pyscf.gto.mole.Mole)

    def __init__(self, system):
        """
        :param system : PySCF object object to be wrapped in this ``PyscfSystem``
        """
        super().__init__()
        if not self.can_wrap(system):
            raise Exception(
                "this class expects pyscf.gto.mole.Mole"
                + "or pyscf.pbc.gto.cell.Cell objects"
            )

        self._system = system
        self._types = self._system.atom_charges().copy()  # dtype=int32
        for atm_i, atomic_type in enumerate(self._types):
            if atomic_type == 0:
                symb = self._system.atom_symbol(atm_i)
                chg = pyscf.data.elements.index(symb)
                self._types[atm_i] = -chg
        if hasattr(pyscf, "pbc"):
            self.is_periodic = isinstance(self._system, pyscf.pbc.gto.cell.Cell)
        else:
            self.is_periodic = False

    def size(self):
        return self._system.natm

    def types(self):
        return self._types

    def positions(self):
        return pyscf.data.nist.BOHR * self._system.atom_coords()

    def cell(self):
        if self.is_periodic:
            cell = self._system.a
            if self._system.unit[0].lower() == "a":
                # assume angströms, we are good to go
                return cell
            else:
                # assume bohrs, correct this
                return pyscf.data.nist.BOHR * np.asarray(cell)
        else:
            return np.zeros((3, 3), float)

    def compute_neighbors(self, cutoff):
        raise Exception("pyscf systems can only be used with 'use_native_system=True'")

    def pairs(self):
        raise Exception("pyscf systems can only be used with 'use_native_system=True'")

    def pairs_containing(self, atom):
        raise Exception("pyscf systems can only be used with 'use_native_system=True'")



"""
This module contains utilities to convert cartesian TensorMap to spherical and
respectively.
"""

from typing import List, Optional

import numpy as np

from . import _coefficients, _dispatch
from ._backend import (
    Array,
    Labels,
    TensorBlock,
    TensorMap,
    TorchTensor,
    torch_jit_is_scripting,
)


def cartesian_to_spherical(
    tensor: TensorMap,
    components: List[str],
    keep_l_in_keys: Optional[bool] = None,
    remove_blocks_threshold: Optional[float] = 1e-9,
    cg_backend: Optional[str] = None,
    cg_coefficients: Optional[TensorMap] = None,
) -> TensorMap:
    """
    Transform a ``tensor`` of arbitrary rank from cartesian form to a spherical form.

    Starting from a tensor on a basis of product of cartesian coordinates, this function
    computes the same tensor using a basis of spherical harmonics ``Y^M_L``. For
    example, a rank 1 tensor with a single "xyz" component would be represented as a
    single L=1 spherical harmonic; while a rank 5 tensor using a product basis ``ℝ^3 ⊗
    ℝ^3 ⊗ ℝ^3 ⊗ ℝ^3 ⊗ ℝ^3`` would require multiple blocks up to L=5 spherical harmonics.

    A single :py:class:`TensorBlock` in the input might correspond to multiple
    :py:class:`TensorBlock` in the output. The output keys will contain all the
    dimensions of the input keys, plus ``o3_lambda`` (indicating the spherical harmonics
    degree) and ``o3_sigma`` (indicating that this block is a proper- or improper tensor
    with ``+1`` and ``-1`` respectively). If ``keep_l_in_keys`` is ``True`` or if the
    input tensor is a tensor of rank 3 or more, the keys will also contain multiple
    ``l_{i}`` and  ``k_{i}`` dimensions, which indicate which angular momenta have been
    coupled together in which order to get this block.

    ``components`` specifies which ones of the components of the input
    :py:class:`TensorMap` should be transformed from cartesian to spherical. All these
    components will be replaced in the output by a single ``o3_mu`` component,
    corresponding to the spherical harmonics ``M``.

    By default, symmetric tensors will only contain blocks corresponding to
    ``o3_sigma=1``. This is achieved by checking the norm of the blocks after the full
    calculation; and dropping any block with a norm below ``remove_blocks_epsilon``. To
    keep all blocks regardless of their norm, you can set
    ``remove_blocks_epsilon=None``.

    :param tensor: input tensor, using a cartesian product basis
    :param components: components of the input tensor to transform into spherical
        components
    :param keep_l_in_keys: should the output contains the values of angular momenta that
        were combined together? This defaults to ``False`` for rank 1 and 2 tensors,
        and ``True`` for all other tensors.

        Keys named ``l_{i}`` correspond to the input ``components``, with ``l_1`` being
        the last entry in ``components`` and ``l_N`` the first one. Keys named ``k_{i}``
        correspond to intermediary spherical components created during the calculation,
        i.e. a ``k_{i}`` used to be ``o3_lambda``.
    :param remove_blocks_threshold: Numerical tolerance to use when determining if a
        block's norm is zero or not. Blocks with zero norm will be excluded from the
        output. Set this to ``None`` to keep all blocks in the output.
    :param cg_backend: Backend to use for Clebsch-Gordan calculations. This can be
        ``"python-dense"`` or ``"python-sparse"`` for dense or sparse operations
        respectively. If ``None``, this is automatically determined.
    :param cg_coefficients: Cache containing Clebsch-Gordan coefficients. This is
        optional except when using this function from TorchScript. The coefficients
        should be computed with :py:func:`calculate_cg_coefficients`, using the same
        ``cg_backend`` as this function.

    :return: :py:class:`TensorMap` containing spherical components instead of cartesian
        components.
    """
    if len(tensor) == 0 or len(components) == 0:
        # nothing to do
        return tensor

    if not isinstance(components, list):
        raise TypeError(f"`components` should be a list, got {type(components)}")

    if keep_l_in_keys is None:
        if len(components) < 3:
            keep_l_in_keys = False
        else:
            keep_l_in_keys = True

    axes_to_convert: List[int] = []
    all_component_names = tensor.component_names
    for component in components:
        if component in all_component_names:
            idx = all_component_names.index(component)
            axes_to_convert.append(idx + 1)
        else:
            raise ValueError(f"'{component}' is not part of this tensor components")

    for key, block in tensor.items():
        for idx in axes_to_convert:
            values_list = _dispatch.to_int_list(block.components[idx - 1].values[:, 0])
            if values_list != [0, 1, 2]:
                name = block.components[idx - 1].names[0]
                raise ValueError(
                    f"component '{name}' in block for {key.print()} should have "
                    f"[0, 1, 2] as values, got {values_list} instead"
                )

    # we need components to be consecutive
    if list(range(axes_to_convert[0], axes_to_convert[-1] + 1)) != axes_to_convert:
        raise ValueError(
            f"this function only supports consecutive components, {components} are not"
        )

    key_names = tensor.keys.names
    if "o3_lambda" in key_names:
        raise ValueError(
            "this tensor already has an `o3_lambda` key, "
            "is it already in spherical form?"
        )

    for i in range(len(components)):
        if f"l_{i}" in key_names:
            raise ValueError(
                f"this tensor already has an `l_{i}` key, "
                "is it already in spherical form?"
            )

    tensor_rank = len(components)
    if tensor_rank > 2 and not keep_l_in_keys:
        raise ValueError(
            "`keep_l_in_keys` must be `True` for tensors of rank 3 and above"
        )

    if isinstance(tensor.block(0).values, TorchTensor):
        arrays_backend = "torch"
        values = tensor.block(0).values
        dtype = values.dtype
        device = values.device
    elif isinstance(tensor.block(0).values, np.ndarray):
        arrays_backend = "numpy"
        values = tensor.block(0).values
        dtype = values.dtype
        device = "cpu"
    else:
        raise TypeError(
            f"unknown array type in tensor ({type(tensor.block(0).values)}), "
            "only numpy and torch are supported"
        )

    if cg_backend is None:
        # TODO: benchmark & change the default?
        if arrays_backend == "torch":
            cg_backend = "python-dense"
        else:
            cg_backend = "python-sparse"

    new_component_names: List[str] = []
    for idx in range(len(tensor.component_names)):
        if idx + 1 in axes_to_convert:
            if len(axes_to_convert) == 1:
                new_component_names.append("o3_mu")
            else:
                new_component_names.append(f"__internal_to_convert_{idx}")
        else:
            new_component_names.append("")

    # Step 1: transform xyz dimensions to o3_lambda=1 dimensions
    # This is done with `roll`, since (y, z, x) is the same as m = (-1, 0, 1)
    new_blocks: List[TensorBlock] = []
    for block in tensor.blocks():
        values = _dispatch.roll(
            block.values,
            shifts=[-1] * len(axes_to_convert),
            axis=axes_to_convert,
        )

        new_components: List[Labels] = []
        for idx, component in enumerate(block.components):
            if idx + 1 in axes_to_convert:
                new_components.append(
                    Labels(
                        names=[new_component_names[idx]],
                        values=_dispatch.int_array_like(
                            [[-1], [0], [1]], component.values
                        ),
                    )
                )
            else:
                new_components.append(component)

        new_blocks.append(
            TensorBlock(values, block.samples, new_components, block.properties)
        )

    if tensor_rank == 1:
        new_keys = tensor.keys
        # we are done, add o3_lambda/o3_sigma/l_1 to the keys & return
        if keep_l_in_keys:
            new_keys = new_keys.insert(
                0,
                "l_1",
                _dispatch.int_array_like([1] * len(new_keys), new_keys.values),
            )

        new_keys = new_keys.insert(
            0,
            "o3_sigma",
            _dispatch.int_array_like([1] * len(tensor.keys), tensor.keys.values),
        )

        new_keys = new_keys.insert(
            0,
            "o3_lambda",
            _dispatch.int_array_like([1] * len(tensor.keys), tensor.keys.values),
        )

        return TensorMap(new_keys, new_blocks)

    # Step 2: if there is more than one dimension, couple them with CG coefficients
    #
    # We start from an array of shape [..., 3, 3, 3, 3, 3, ...] with as many 3 as
    # `len(components)`. Then we iteratively combine the two rightmost components into
    # as many new lambda/mu entries as required, until there is only one component left.
    # Each step will create multiple blocks (corresponding to the different o3_lambda
    # created by combining two o3_lambda=1 terms), that might on their turn create more
    # blocks if more combinations are required.
    #
    # For example, with a rank 3 tensor we go through the following:
    #
    # - Step 1: [..., 3, 3, 3, ...] => [..., 3, 1, ...] (o3_lambda=0, o3_sigma=+1)
    #                               => [..., 3, 3, ...] (o3_lambda=1, o3_sigma=-1)
    #                               => [..., 3, 5, ...] (o3_lambda=2, o3_sigma=+1)
    #
    # - Step 2: [..., 3, 1, ...] => [..., 3, ...] (o3_lambda=1, o3_sigma=+1)
    #           [..., 3, 3, ...] => [..., 1, ...] (o3_lambda=0, o3_sigma=-1)
    #                            => [..., 3, ...] (o3_lambda=1, o3_sigma=+1)
    #                            => [..., 5, ...] (o3_lambda=2, o3_sigma=-1)
    #           [..., 3, 5, ...] => [..., 3, ...] (o3_lambda=1, o3_sigma=+1)
    #                            => [..., 5, ...] (o3_lambda=2, o3_sigma=-1)
    #                            => [..., 7, ...] (o3_lambda=3, o3_sigma=+1)

    # Use the rolled block values as the starting point for higher-rank tensors
    tensor = TensorMap(tensor.keys, new_blocks)

    if cg_coefficients is None:
        if torch_jit_is_scripting():
            raise ValueError(
                "in TorchScript mode, `cg_coefficients` must be pre-computed "
                "and given to this function explicitly"
            )
        else:
            cg_coefficients = _coefficients.calculate_cg_coefficients(
                lambda_max=len(axes_to_convert),
                cg_backend=cg_backend,
                arrays_backend=arrays_backend,
                dtype=dtype,
                device=device,
            )

    iteration_index = 0
    while len(axes_to_convert) > 1:
        tensor = _do_coupling(
            tensor=tensor,
            component_1=axes_to_convert[-2] - 1,
            component_2=axes_to_convert[-1] - 1,
            cg_coefficients=cg_coefficients,
            cg_backend=cg_backend,
            keep_l_in_keys=keep_l_in_keys,
            iteration_index=iteration_index,
        )

        axes_to_convert.pop()
        iteration_index += 1

    if remove_blocks_threshold is None:
        return tensor

    # Step 3: for symmetry reasons, some of the blocks will be zero everywhere (for
    # example o3_sigma=-1 blocks if the input tensor is fully symmetric). If the user
    # gave us a threshold, we remove all blocks with a norm below this threshold.
    new_keys_values: List[Array] = []
    new_blocks: List[TensorBlock] = []
    for key_idx, block in enumerate(tensor.blocks()):
        key = tensor.keys.entry(key_idx)
        values = block.values.reshape(-1, 1)
        norm = values.T @ values
        if norm > remove_blocks_threshold:
            new_keys_values.append(key.values.reshape(1, -1))
            new_blocks.append(
                TensorBlock(
                    values=block.values,
                    samples=block.samples,
                    components=block.components,
                    properties=block.properties,
                )
            )

    return TensorMap(
        Labels(tensor.keys.names, _dispatch.concatenate(new_keys_values, axis=0)),
        new_blocks,
    )


def _do_coupling(
    tensor: TensorMap,
    component_1: int,
    component_2: int,
    keep_l_in_keys: bool,
    iteration_index: int,
    cg_coefficients: TensorMap,
    cg_backend: str,
) -> TensorMap:
    """
    Go from an uncoupled product basis that behave like a product of spherical harmonics
    to a coupled basis that behaves like a single spherical harmonic.

    This function takes in a :py:class:`TensorMap` where two of the components
    (indicated by ``component_1`` and ``component_2``) behave like spherical harmonics
    ``Y^m1_l1`` and ``Y^m2_l2``, and project it onto a single spherical harmonic
    ``Y^M_L``. This transformation uses the following relation:

    ``|L M> = |l1 l2 L M> = \\sum_{m1 m2} <l1 m1 l2 m2|L M> |l1 m1> |l2 m2>``

    where ``<l1 m1 l2 m2|L M>`` are Clebsch-Gordan coefficients.

    The output will contain many blocks for each block in the input, matching all the
    different ``L`` (called ``o3_lambda`` in the code) required to do a full projection.

    This process can be iterated: a multi-dimensional array that is the product of many
    ``Y^m_l`` can be turned into a set of multiple terms transforming as a single
    ``Y^M_L``.

    :param tensor: input :py:class:`TensorMap`
    :param components_1: first component of the ``tensor`` behaving like spherical
        harmonics
    :param components_2: second component of the ``tensor`` behaving like spherical
        harmonics
    :param keep_l_in_keys: whether ``l1`` and ``l2`` (the original spherical harmonic
        degrees) should be kept in the keys. This can be useful to undo this
        transformation (or even required if there is more than one path to get to a
        given value for ``o3_lambda``)
    :param iteration_index: when iterating the coupling, this should be the number of
        iterations already done (i.e. the number of time this function has been called)
    :param cg_coefficients: pre-computed set of Clebsch-Gordan coefficients
    :param cg_backend: which backend to use for the calculations

    :return: :py:class:`TensorMap` using the coupled basis. This will contain the same
        keys as the input ``tensor``, plus ``o3_lambda``. The components in positions
        ``components_1`` and ``components_2`` will be replaced by a single ``o3_mu``
        component.
    """
    assert component_2 == component_1 + 1

    new_keys = tensor.keys

    if "o3_lambda" in tensor.keys.names:
        old_sigmas = new_keys.column("o3_sigma")
        new_keys = new_keys.remove("o3_sigma")
    else:
        old_sigmas = _dispatch.int_array_like([1] * len(new_keys), new_keys.values)

    if keep_l_in_keys:
        array_of_ones = _dispatch.int_array_like([1] * len(new_keys), new_keys.values)
        if "o3_lambda" in tensor.keys.names:
            assert iteration_index > 0
            new_keys = new_keys.rename("o3_lambda", f"k_{iteration_index}")
            new_keys = new_keys.insert(0, f"l_{iteration_index + 2}", array_of_ones)
        else:
            assert iteration_index == 0
            new_keys = new_keys.insert(0, "l_1", array_of_ones)
            new_keys = new_keys.insert(0, "l_2", array_of_ones)

    new_keys_values: List[List[int]] = []
    new_blocks: List[TensorBlock] = []
    for key_idx, block in enumerate(tensor.blocks()):
        key = new_keys.entry(key_idx)
        old_sigma = int(old_sigmas[key_idx])

        # get l1, l2 from the block's shape
        block_shape = block.values.shape
        l1 = (block_shape[component_1 + 1] - 1) // 2
        l2 = (block_shape[component_2 + 1] - 1) // 2

        # reshape the values to look like (n_s, 2*l1 + 1, 2*l2 + 1, n_p)
        shape_before = 1
        for axis in range(component_1 + 1):
            shape_before *= block_shape[axis]

        shape_after = 1
        for axis in range(component_2 + 2, len(block_shape)):
            shape_after *= block_shape[axis]

        array = block.values.reshape(
            shape_before,
            block.values.shape[component_1 + 1],
            block.values.shape[component_2 + 1],
            shape_after,
        )

        # generate the set of o3_lambda to compute
        o3_lambdas = list(range(max(l1, l2) - min(l1, l2), (l1 + l2) + 1))

        # actual calculation
        outputs = _coefficients.cg_couple(
            array, o3_lambdas, cg_coefficients, cg_backend
        )

        # create one block for each output of `cg_couple`
        for o3_lambda, values in zip(o3_lambdas, outputs):
            o3_sigma = int(old_sigma * (-1) ** (l1 + l2 + o3_lambda))
            new_keys_values.append(
                [o3_lambda, o3_sigma] + _dispatch.to_int_list(key.values)
            )

            new_shape = list(block.values.shape)
            new_shape.pop(component_2 + 1)
            new_shape[component_1 + 1] = 2 * o3_lambda + 1

            new_components = block.components
            new_components.pop(component_2)
            new_components[component_1] = Labels(
                "o3_mu",
                _dispatch.int_array_like(
                    [[mu] for mu in range(-o3_lambda, o3_lambda + 1)], new_keys.values
                ),
            )

            new_blocks.append(
                TensorBlock(
                    values.reshape(new_shape),
                    samples=block.samples,
                    components=new_components,
                    properties=block.properties,
                )
            )

    new_keys = Labels(
        ["o3_lambda", "o3_sigma"] + new_keys.names,
        _dispatch.int_array_like(new_keys_values, new_keys.values),
    )
    return TensorMap(new_keys, new_blocks)



from typing import Any, Union

import numpy as np
from metatensor import (  # noqa F401
    Labels,
    LabelsEntry,
    TensorBlock,
    TensorMap,
    operations,
)

from ..calculator_base import CalculatorBase  # noqa F401
from ..systems import IntoSystem  # noqa F401


def torch_jit_is_scripting():
    return False


def torch_jit_export(func):
    return func


def is_labels(obj: Any):
    return isinstance(obj, Labels)


try:
    from torch import ScriptClass as TorchScriptClass
    from torch import Tensor as TorchTensor
    from torch import device as TorchDevice
    from torch import dtype as TorchDType
    from torch.nn import Module as TorchModule
except ImportError:

    class TorchTensor:
        pass

    class TorchModule:
        def __call__(self, *arg, **kwargs):
            return self.forward(*arg, **kwargs)

    class TorchScriptClass:
        pass

    class TorchDevice:
        pass

    class TorchDType:
        pass


Array = Union[np.ndarray, TorchTensor]
DType = Union[np.dtype, TorchDType]
Device = Union[str, TorchDevice]

BACKEND_IS_METATENSOR_TORCH = False



from .ase import HAVE_ASE, AseSystem
from .base import SystemBase
from .chemfiles import HAVE_CHEMFILES, ChemfilesSystem
from .pyscf import HAVE_PYSCF, PyscfSystem


class IntoSystem:
    """
    Possible types that can be used as a featomic System.
    """

    def __init__(self):
        raise ValueError(
            "this class is for documentation purposes only and should not "
            "be instantiated"
        )


if HAVE_ASE:
    IntoSystem.__doc__ += """
    - `ase.Atoms`_: the Atomistic Simulation Environment (ASE) Atoms class. The
      ASE neighbor list is used to find neighbors.

    .. _ase.Atoms: https://wiki.fysik.dtu.dk/ase/ase/atoms.html
    """

if HAVE_CHEMFILES:
    IntoSystem.__doc__ += """
    - `chemfiles.Frame`_: chemfiles' Frame type. There is no associated neighbor
      list implementation, the system will only be usable with
      ``use_native_system=True``

    .. _chemfiles.Frame: http://chemfiles.org/chemfiles.py/latest/reference/frame.html
    """

if HAVE_PYSCF:
    IntoSystem.__doc__ += """
    - `pyscf.gto.mole.Mole`_ or `pyscf.pbc.gto.cell.Cell`_: pyscf' Frame type.
       There is no associated neighbor list implementation, the system will only
       be usable with ``use_native_system=True``

    .. _pyscf.gto.mole.Mole: https://pyscf.org/user/gto.html
    .. _pyscf.pbc.gto.cell.Cell: https://pyscf.org/user/pbc/gto.html
    """


def wrap_system(system: IntoSystem) -> SystemBase:
    """Wrap different systems implementation into the right class.

    This function is automatically called on all systems passed to
    :py:func:`featomic.calculators.CalculatorBase.compute`. This function makes
    different systems compatible with featomic.

    The supported system types are documented in the
    py:class:`featomic.IntoSystem` class. If ``system`` is already a subclass
    of :py:class:`featomic.SystemBase`, it is returned as-is.

    :param system: external system to wrap

    :returns: a specialized instance of :py:class:`featomic.SystemBase`
    """
    if isinstance(system, SystemBase):
        return system

    if HAVE_ASE and AseSystem.can_wrap(system):
        return AseSystem(system)

    if HAVE_PYSCF and PyscfSystem.can_wrap(system):
        return PyscfSystem(system)

    if HAVE_CHEMFILES and ChemfilesSystem.can_wrap(system):
        return ChemfilesSystem(system)

    raise TypeError(f"unknown system type: {type(system)}")



import glob
import os
import subprocess
import sys

import packaging
from setuptools import Extension, setup
from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist
from wheel.bdist_wheel import bdist_wheel


ROOT = os.path.realpath(os.path.dirname(__file__))
FEATOMIC_SRC = os.path.realpath(os.path.join(ROOT, "..", "..", "featomic"))

FEATOMIC_BUILD_TYPE = os.environ.get("FEATOMIC_BUILD_TYPE", "release")
if FEATOMIC_BUILD_TYPE not in ["debug", "release"]:
    raise Exception(
        f"invalid build type passed: '{FEATOMIC_BUILD_TYPE}',"
        "expected 'debug' or 'release'"
    )

# The code for featomic-torch needs to live in `featomic_torch` directory until
# https://github.com/pypa/pip/issues/13093 is fixed.
FEATOMIC_TORCH_SRC = os.path.realpath(os.path.join(ROOT, "..", "featomic_torch"))


class universal_wheel(bdist_wheel):
    # When building the wheel, the `wheel` package assumes that if we have a
    # binary extension then we are linking to `libpython.so`; and thus the wheel
    # is only usable with a single python version. This is not the case for
    # here, and the wheel will be compatible with any Python >=3. This is
    # tracked in https://github.com/pypa/wheel/issues/185, but until then we
    # manually override the wheel tag.
    def get_tag(self):
        tag = bdist_wheel.get_tag(self)
        # tag[2:] contains the os/arch tags, we want to keep them
        return ("py3", "none") + tag[2:]


class cmake_ext(build_ext):
    """Build the native library using cmake."""

    def run(self):
        """Run cmake build and install the resulting library."""
        import metatensor

        source_dir = FEATOMIC_SRC
        build_dir = os.path.join(ROOT, "build", "cmake-build")
        install_dir = os.path.join(os.path.realpath(self.build_lib), "featomic")

        try:
            os.mkdir(build_dir)
        except OSError:
            pass

        cmake_options = [
            f"-DCMAKE_INSTALL_PREFIX={install_dir}",
            "-DCMAKE_INSTALL_LIBDIR=lib",
            f"-DCMAKE_BUILD_TYPE={FEATOMIC_BUILD_TYPE}",
            f"-DCMAKE_PREFIX_PATH={metatensor.utils.cmake_prefix_path}",
            "-DFEATOMIC_INSTALL_BOTH_STATIC_SHARED=OFF",
            "-DBUILD_SHARED_LIBS=ON",
            "-DEXTRA_RUST_FLAGS=-Cstrip=symbols",
        ]

        if "CARGO" in os.environ:
            cmake_options.append(f"-DCARGO_EXE={os.environ['CARGO']}")

        # Handle cross-compilation by detecting cibuildwheels environnement
        # variables
        if sys.platform.startswith("darwin"):
            # ARCHFLAGS is set by cibuildwheels
            ARCHFLAGS = os.environ.get("ARCHFLAGS")
            if ARCHFLAGS is not None:
                archs = filter(
                    lambda u: bool(u),
                    ARCHFLAGS.strip().split("-arch "),
                )
                archs = list(archs)
                assert len(archs) == 1
                arch = archs[0].strip()

                if arch == "x86_64":
                    cmake_options.append("-DRUST_BUILD_TARGET=x86_64-apple-darwin")
                elif arch == "arm64":
                    cmake_options.append("-DRUST_BUILD_TARGET=aarch64-apple-darwin")
                else:
                    raise ValueError(f"unknown arch: {arch}")

        elif sys.platform.startswith("linux"):
            # we set RUST_BUILD_TARGET in our custom docker image
            RUST_BUILD_TARGET = os.environ.get("RUST_BUILD_TARGET")
            if RUST_BUILD_TARGET is not None:
                cmake_options.append(f"-DRUST_BUILD_TARGET={RUST_BUILD_TARGET}")

        elif sys.platform.startswith("win32"):
            # CARGO_BUILD_TARGET is set by cibuildwheels
            CARGO_BUILD_TARGET = os.environ.get("CARGO_BUILD_TARGET")
            if CARGO_BUILD_TARGET is not None:
                cmake_options.append(f"-DRUST_BUILD_TARGET={CARGO_BUILD_TARGET}")

        else:
            raise ValueError(f"unknown platform: {sys.platform}")

        subprocess.run(
            ["cmake", source_dir, *cmake_options],
            cwd=build_dir,
            check=True,
        )
        subprocess.run(
            ["cmake", "--build", build_dir, "--parallel", "--target", "install"],
            check=True,
        )


class bdist_egg_disabled(bdist_egg):
    """Disabled version of bdist_egg

    Prevents setup.py install performing setuptools' default easy_install,
    which it should never ever do.
    """

    def run(self):
        sys.exit(
            "Aborting implicit building of eggs. "
            + "Use `pip install .` or `python setup.py bdist_wheel && pip "
            + "uninstall featomic -y && pip install dist/featomic-*.whl` "
            + "to install from source."
        )


class sdist_generate_data(sdist):
    """
    Create a sdist with an additional generated files:
        - `git_version_info`
        - `featomic-cxx-*.tar.gz`
    """

    def run(self):
        n_commits, git_hash = git_version_info()
        with open("git_version_info", "w") as fd:
            fd.write(f"{n_commits}\n{git_hash}\n")

        generate_cxx_tar()

        # run original sdist
        super().run()

        os.unlink("git_version_info")
        for path in glob.glob("featomic-cxx-*.tar.gz"):
            os.unlink(path)


def generate_cxx_tar():
    script = os.path.join(ROOT, "..", "..", "scripts", "package-featomic.sh")
    assert os.path.exists(script)

    try:
        output = subprocess.run(
            ["bash", "--version"],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            encoding="utf8",
        )
    except Exception as e:
        raise RuntimeError("could not run `bash`, is it installed?") from e

    output = subprocess.run(
        ["bash", script, os.getcwd()],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        encoding="utf8",
    )
    if output.returncode != 0:
        stderr = output.stderr
        stdout = output.stdout
        raise RuntimeError(
            "failed to collect C++ sources for Python sdist\n"
            f"stdout:\n {stdout}\n\nstderr:\n {stderr}"
        )


def get_rust_version():
    # read version from Cargo.toml
    with open(os.path.join(FEATOMIC_SRC, "Cargo.toml")) as fd:
        for line in fd:
            if line.startswith("version"):
                _, version = line.split(" = ")
                # remove quotes
                version = version[1:-2]
                # take the first version in the file, this should be the right
                # version
                break

    return version


def git_version_info():
    """
    If git is available and we are building from a checkout, get the number of commits
    since the last tag & full hash of the code. Otherwise, this always returns (0, "").
    """
    TAG_PREFIX = "featomic-v"

    if os.path.exists("git_version_info"):
        # we are building from a sdist, without git available, but the git
        # version was recorded in the `git_version_info` file
        with open("git_version_info") as fd:
            n_commits = int(fd.readline().strip())
            git_hash = fd.readline().strip()
    else:
        script = os.path.join(ROOT, "..", "..", "scripts", "git-version-info.py")
        assert os.path.exists(script)

        output = subprocess.run(
            [sys.executable, script, TAG_PREFIX],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            encoding="utf8",
        )

        if output.returncode != 0:
            raise Exception(
                "failed to get git version info.\n"
                f"stdout: {output.stdout}\n"
                f"stderr: {output.stderr}\n"
            )
        elif output.stderr:
            print(output.stderr, file=sys.stderr)
            n_commits = 0
            git_hash = ""
        else:
            lines = output.stdout.splitlines()
            n_commits = int(lines[0].strip())
            git_hash = lines[1].strip()

    return n_commits, git_hash


def create_version_number(version):
    version = packaging.version.parse(version)

    n_commits, git_hash = git_version_info()
    if n_commits != 0:
        # `n_commits` will be non zero only if we have commits since the last tag. This
        # mean we are in a pre-release of the next version. So we increase either the
        # minor version number or the release candidate number (if we are closing up on
        # a release)
        if version.pre is not None:
            assert version.pre[0] == "rc"
            pre = ("rc", version.pre[1] + 1)
            release = version.release
        else:
            major, minor, patch = version.release
            release = (major, minor + 1, 0)
            pre = None

        # this is using a private API which is intended to become public soon:
        # https://github.com/pypa/packaging/pull/698. In the mean time we'll
        # use this
        version._version = version._version._replace(release=release)
        version._version = version._version._replace(pre=pre)
        version._version = version._version._replace(dev=("dev", n_commits))
        version._version = version._version._replace(local=(git_hash,))

    return str(version)


if __name__ == "__main__":
    if not os.path.exists(FEATOMIC_SRC):
        # we are building from a sdist, which should include featomic Rust
        # sources as a tarball
        tarballs = glob.glob(os.path.join(ROOT, "featomic-*.tar.gz"))

        if not len(tarballs) == 1:
            raise RuntimeError(
                "expected a single 'featomic-*.tar.gz' file containing "
                "featomic Rust sources. remove all files and re-run "
                "scripts/package-featomic.sh"
            )

        FEATOMIC_SRC = os.path.realpath(tarballs[0])
        subprocess.run(
            ["cmake", "-E", "tar", "xf", FEATOMIC_SRC],
            cwd=ROOT,
            check=True,
        )

        FEATOMIC_SRC = ".".join(FEATOMIC_SRC.split(".")[:-2])

    with open(os.path.join(ROOT, "AUTHORS")) as fd:
        authors = fd.read().splitlines()

    extras_require = {}

    # when packaging a sdist for release, we should never use local dependencies
    FEATOMIC_NO_LOCAL_DEPS = os.environ.get("FEATOMIC_NO_LOCAL_DEPS", "0") == "1"
    if not FEATOMIC_NO_LOCAL_DEPS and os.path.exists(FEATOMIC_TORCH_SRC):
        # we are building from a git checkout
        extras_require["torch"] = f"featomic-torch @ file://{FEATOMIC_TORCH_SRC}"
    else:
        # we are building from a sdist/installing from a wheel
        extras_require["torch"] = "featomic-torch"

    setup(
        version=create_version_number(get_rust_version()),
        author=", ".join(authors),
        extras_require=extras_require,
        ext_modules=[
            # only declare the extension, it is built & copied as required by cmake
            # in the build_ext command
            Extension(name="featomic", sources=[]),
        ],
        cmdclass={
            "build_ext": cmake_ext,
            "bdist_egg": bdist_egg if "bdist_egg" in sys.argv else bdist_egg_disabled,
            "bdist_wheel": universal_wheel,
            "sdist": sdist_generate_data,
        },
        package_data={
            "featomic": [
                "featomic/lib/*",
                "featomic/include/*",
            ]
        },
    )



import json
from math import sqrt
from typing import List, Optional, Union

from . import _dispatch
from ._backend import (
    CalculatorBase,
    IntoSystem,
    Labels,
    TensorBlock,
    TensorMap,
    TorchModule,
)


class PowerSpectrum(TorchModule):
    r"""General power spectrum of one or of two calculators.

    If ``calculator_2`` is provided, the invariants :math:`p_{nl}` are generated by
    taking quadratic combinations of ``calculator_1``'s spherical expansion
    :math:`\rho_{nlm}` and ``calculator_2``'s spherical expansion :math:`\nu_{nlm}`
    according to `Bartók et. al
    <https://journals.aps.org/prb/abstract/10.1103/PhysRevB.87.184115>`_.

    .. math::
        p_{nl} = \rho_{nlm}^\dagger \cdot \nu_{nlm}

    where we use the Einstein summation convention. If gradients are present the
    invariants of those are constructed as

    .. math::
        \nabla p_{nl} = \nabla \rho_{nlm}^\dagger \cdot \nu_{nlm} +
                        \rho_{nlm}^\dagger \cdot \nabla \nu_{nlm}

    .. note::
        Currently only supports gradients with respect to ``positions``.

    If ``calculator_2=None`` invariants are generated by combining the coefficients of
    the spherical expansion of ``calculator_1``. The spherical expansions given as input
    can only be :py:class:`featomic.SphericalExpansion` or
    :py:class:`featomic.LodeSphericalExpansion`.

    Example
    -------
    As an example we calculate the power spectrum for a short range (sr) spherical
    expansion and a long-range (lr) LODE spherical expansion for a NaCl crystal.

    >>> import featomic
    >>> import ase

    Construct the NaCl crystal

    >>> atoms = ase.Atoms(
    ...     symbols="NaCl",
    ...     positions=[[0, 0, 0], [0.5, 0.5, 0.5]],
    ...     pbc=True,
    ...     cell=[1, 1, 1],
    ... )

    Define the hyper parameters for the short-range spherical expansion

    >>> sr_hypers = {
    ...     "cutoff": {
    ...         "radius": 1.0,
    ...         "smoothing": {"type": "ShiftedCosine", "width": 0.5},
    ...     },
    ...     "density": {
    ...         "type": "Gaussian",
    ...         "width": 0.3,
    ...     },
    ...     "basis": {
    ...         "type": "TensorProduct",
    ...         "max_angular": 2,
    ...         "radial": {"type": "Gto", "max_radial": 5},
    ...     },
    ... }

    Define the hyper parameters for the long-range LODE spherical expansion from the
    hyper parameters of the short-range spherical expansion

    >>> lr_hypers = {
    ...     "density": {
    ...         "type": "SmearedPowerLaw",
    ...         "smearing": 0.3,
    ...         "exponent": 1,
    ...     },
    ...     "basis": {
    ...         "type": "TensorProduct",
    ...         "max_angular": 2,
    ...         "radial": {"type": "Gto", "max_radial": 5, "radius": 1.0},
    ...     },
    ... }

    Construct the calculators

    >>> sr_calculator = featomic.SphericalExpansion(**sr_hypers)
    >>> lr_calculator = featomic.LodeSphericalExpansion(**lr_hypers)

    Construct the power spectrum calculators and compute the spherical expansion

    >>> calculator = featomic.clebsch_gordan.PowerSpectrum(sr_calculator, lr_calculator)
    >>> power_spectrum = calculator.compute(atoms)

    The resulting invariants are stored as :py:class:`metatensor.TensorMap` as for any
    other calculator

    >>> power_spectrum.keys
    Labels(
        center_type
            11
            17
    )
    >>> power_spectrum[0]
    TensorBlock
        samples (1): ['system', 'atom']
        components (): []
        properties (432): ['l', 'neighbor_1_type', 'n_1', 'neighbor_2_type', 'n_2']
        gradients: None


    .. seealso::
        If you are interested in the SOAP power spectrum you can the use the
        faster :py:class:`featomic.SoapPowerSpectrum`. For an equivariant version of
        this calculator, computing the power spectrum for covariant as well as
        invariant blocks, see :py:class:`featomic.EquivariantPowerSpectrum`.
    """

    def __init__(
        self,
        calculator_1: CalculatorBase,
        calculator_2: Optional[CalculatorBase] = None,
        neighbor_types: Optional[List[int]] = None,
    ):
        """
        Constructs the power spectrum calculator.

        :param calculator_1: first calculator that computes a density descriptor, either
            a :py:class:`featomic.SphericalExpansion` or
            :py:class:`featomic.LodeSphericalExpansion`.
        :param calculator_2: optional second calculator that computes a density
            descriptor, either a :py:class:`featomic.SphericalExpansion` or
            :py:class:`featomic.LodeSphericalExpansion`. If ``None``, the power spectrum
            is computed as the auto-correlation of the first calculator. Defaults to
            ``None``.
        :param neighbor_types: List of ``"neighbor_type"`` to use in the properties of
            the output. This option might be useful when running the calculation on
            subset of a whole dataset and trying to join along the ``sample`` dimension
            after the calculation. If ``None``, blocks are filled with
            ``"neighbor_type"`` found in the systems.
        """
        super().__init__()
        self.calculator_1 = calculator_1
        self.calculator_2 = calculator_2
        self.neighbor_types = neighbor_types

        supported_calculators = ["lode_spherical_expansion", "spherical_expansion"]

        if self.calculator_1.c_name not in supported_calculators:
            raise ValueError(
                f"Only [{', '.join(supported_calculators)}] are supported for "
                f"`calculator_1`, got '{self.calculator_1.c_name}'"
            )

        if self.calculator_2 is not None:
            if self.calculator_2.c_name not in supported_calculators:
                raise ValueError(
                    f"Only [{', '.join(supported_calculators)}] are supported for "
                    f"`calculator_2`, got '{self.calculator_2.c_name}'"
                )

            parameters_1 = json.loads(calculator_1.parameters)
            parameters_2 = json.loads(calculator_2.parameters)
            if parameters_1["basis"]["type"] != "TensorProduct":
                raise ValueError(
                    "only 'TensorProduct' basis is supported for calculator_1"
                )

            if parameters_2["basis"]["type"] != "TensorProduct":
                raise ValueError(
                    "only 'TensorProduct' basis is supported for calculator_2"
                )

            max_angular_1 = parameters_1["basis"]["max_angular"]
            max_angular_2 = parameters_2["basis"]["max_angular"]
            if max_angular_1 != max_angular_2:
                raise ValueError(
                    "'basis.max_angular' must be the same in both calculators, "
                    f"got {max_angular_1} and {max_angular_2}"
                )

    @property
    def name(self):
        """Name of this calculator."""
        return "PowerSpectrum"

    def compute(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        gradients: Optional[List[str]] = None,
        use_native_system: bool = True,
    ) -> TensorMap:
        """Runs a calculation with this calculator on the given ``systems``.

        See :py:func:`featomic.calculators.CalculatorBase.compute()` for details on the
        parameters.

        :raises NotImplementedError: If a spherical expansions contains a gradient with
            respect to an unknwon parameter.
        """
        if gradients is not None:
            for parameter in gradients:
                if parameter != "positions":
                    raise NotImplementedError(
                        "PowerSpectrum currently only supports gradients "
                        "w.r.t. to positions"
                    )

        spherical_expansion_1 = self.calculator_1.compute(
            systems=systems, gradients=gradients, use_native_system=use_native_system
        )

        expected_key_names = [
            "o3_lambda",
            "o3_sigma",
            "center_type",
            "neighbor_type",
        ]

        assert spherical_expansion_1.keys.names == expected_key_names
        assert spherical_expansion_1.property_names == ["n"]

        if self.neighbor_types is None:
            # Fill blocks with `neighbor_type` from ALL blocks. If we don't do this
            # merging blocks along the ``sample`` direction might be not possible.
            array = spherical_expansion_1.keys.column("neighbor_type")
            values = _dispatch.unique(array).reshape(-1, 1)
        else:
            # Take user provided `neighbor_type` list.
            values = _dispatch.list_to_array(
                array=spherical_expansion_1.keys.values,
                data=[[t] for t in self.neighbor_types],
            )

        keys_to_move = Labels(names="neighbor_type", values=values)

        spherical_expansion_1 = spherical_expansion_1.keys_to_properties(keys_to_move)

        if self.calculator_2 is None:
            spherical_expansion_2 = spherical_expansion_1
        else:
            spherical_expansion_2 = self.calculator_2.compute(
                systems=systems,
                gradients=gradients,
                use_native_system=use_native_system,
            )
            assert spherical_expansion_2.keys.names == expected_key_names
            assert spherical_expansion_2.property_names == ["n"]

            if self.neighbor_types is None:
                array = spherical_expansion_2.keys.column("neighbor_type")
                values = _dispatch.unique(array).reshape(-1, 1)
            else:
                values = _dispatch.list_to_array(
                    array=spherical_expansion_2.keys.values,
                    data=[[t] for t in self.neighbor_types],
                )

            keys_to_move = Labels(names="neighbor_type", values=values)

            spherical_expansion_2 = spherical_expansion_2.keys_to_properties(
                keys_to_move
            )

        new_blocks: List[TensorBlock] = []
        new_keys_values: List[List[int]] = []

        for key, block_1 in spherical_expansion_1.items():
            o3_lambda = key["o3_lambda"]
            center_type = key["center_type"]

            # For consistency with a full Clebsch-Gordan product we need to add
            # a `-1^l / sqrt(2 l + 1)` factor to the power spectrum invariants
            factor = (-1) ** o3_lambda / sqrt(2 * o3_lambda + 1)

            # Find that block indices that have the same o3_lambda and center_type
            selection = Labels(
                names=["o3_lambda", "center_type"],
                values=_dispatch.list_to_array(
                    data=[[o3_lambda, center_type]],
                    array=spherical_expansion_1.keys.values,
                ),
            )
            blocks_2 = spherical_expansion_2.blocks(selection)
            for block_2 in blocks_2:
                # Make sure that samples are the same. This should not happen.
                assert block_1.samples == block_2.samples

                properties_1 = block_1.properties
                properties_2 = block_2.properties

                n_keys_dimensions = (
                    properties_1.values.shape[1] + properties_2.values.shape[1]
                )

                new_property_values = _dispatch.empty_like(
                    array=properties_1.values,
                    shape=[
                        properties_1.values.shape[0],
                        properties_2.values.shape[0],
                        n_keys_dimensions,
                    ],
                )

                for i, values_1 in enumerate(properties_1.values):
                    for j, values_2 in enumerate(properties_2.values):
                        new_property_values[i, j, : len(values_1)] = values_1
                        new_property_values[i, j, len(values_1) :] = values_2

                properties = Labels(
                    names=["neighbor_1_type", "n_1", "neighbor_2_type", "n_2"],
                    values=new_property_values.reshape(-1, n_keys_dimensions),
                )

                # Compute the invariants by summation and store the results this is
                # equivalent to an einsum with: ima, imb -> iab
                data = factor * _dispatch.matmul(
                    block_1.values.swapaxes(1, 2), block_2.values
                )

                new_block = TensorBlock(
                    values=data.reshape(data.shape[0], -1),
                    samples=block_1.samples,
                    components=[],
                    properties=properties,
                )

                for parameter in block_1.gradients_list():
                    if parameter == "positions":
                        _positions_gradients(new_block, block_1, block_2, factor)

                new_keys_values.append([o3_lambda, center_type])
                new_blocks.append(new_block)

        new_keys = Labels(
            names=["l", "center_type"],
            values=_dispatch.list_to_array(
                array=spherical_expansion_1.keys.values, data=new_keys_values
            ),
        )

        return TensorMap(new_keys, new_blocks).keys_to_properties("l")

    def forward(
        self,
        systems: Union[IntoSystem, List[IntoSystem]],
        gradients: Optional[List[str]] = None,
        use_native_system: bool = True,
    ) -> TensorMap:
        """
        Calls the :py:meth:`PowerSpectrum.compute` function.

        This is intended for :py:class:`torch.nn.Module` compatibility, and should be
        ignored in pure Python mode.
        """

        return self.compute(
            systems=systems,
            gradients=gradients,
            use_native_system=use_native_system,
        )


def _positions_gradients(
    new_block: TensorBlock, block_1: TensorBlock, block_2: TensorBlock, factor: float
):
    gradient_1 = block_1.gradient("positions")
    gradient_2 = block_2.gradient("positions")

    if len(gradient_1.samples) == 0 or len(gradient_2.samples) == 0:
        gradients_samples = Labels.empty(["sample", "system", "atom"])
        gradient_values = _dispatch.list_to_array(
            array=gradient_1.values, data=[]
        ).reshape(0, 1, len(new_block.properties))
    else:
        # The "sample" dimension in the power spectrum gradient samples do
        # not necessarily matches the "sample" dimension in the spherical
        # expansion gradient samples. We create new samples by creating a
        # union between the two gradient samples.
        (
            gradients_samples,
            grad1_sample_idxs,
            grad2_sample_idxs,
        ) = gradient_1.samples.union_and_mapping(gradient_2.samples)

        gradient_values = _dispatch.zeros_like(
            array=gradient_1.values,
            shape=[gradients_samples.values.shape[0], 3, len(new_block.properties)],
        )

        # the operation below is equivalent to an einsum with: ixma, imb -> ixab
        sample_indices_1 = _dispatch.to_index_array(gradient_1.samples.column("sample"))
        block_2_values = block_2.values[sample_indices_1]
        new_shape = block_2_values.shape[:1] + (-1,) + block_2_values.shape[1:]

        gradient_1_values = factor * _dispatch.matmul(
            gradient_1.values.swapaxes(2, 3),
            block_2_values.reshape(new_shape),
        )

        gradient_values[grad1_sample_idxs] += gradient_1_values.reshape(
            gradient_1.samples.values.shape[0], 3, -1
        )

        # the operation below is equivalent to an einsum with: ima, ixmb -> ixab
        sample_indices_2 = _dispatch.to_index_array(gradient_2.samples.column("sample"))
        block_1_values = block_1.values[sample_indices_2]
        new_shape = block_1_values.shape[:1] + (-1,) + block_1_values.shape[1:]

        gradient_values_2 = factor * _dispatch.matmul(
            block_1_values.reshape(new_shape).swapaxes(2, 3),
            gradient_2.values,
        )

        gradient_values[grad2_sample_idxs] += gradient_values_2.reshape(
            gradient_2.samples.values.shape[0], 3, -1
        )

    gradient = TensorBlock(
        values=gradient_values,
        samples=gradients_samples,
        components=[gradient_1.components[0]],
        properties=new_block.properties,
    )

    new_block.add_gradient("positions", gradient)
